--------------------------------------------
--TreeMatoGraph 
--version 0.2.5 (Beta)
--Last Edited: 08/02/2002
--Copyright (c) 2002 by Borislav Petrov
--bobo@email.archlab.tuwien.ac.at
--------------------------------------------
--New In 0.2.5:
--*Experimental code using Larry's Avguard DLX for pasteBitmap - much faster Panning and Moving when available (R5 only)
--*Fixed crasher when creating maps/materials without a Root Material
--*New Control Behaviors: RMB = Link, RMB+Control = Unlink. All modes snap back to Move when released.

--New In 0.2.4:
--*Added temp. error trap for Unlinking from Raytrace material - doesn't work right now.

--New In 0.2.3:
--*Added 2 additional link modes - Straight Horizontal and Vertical.

--New In 0.2.2:
--*Added Material and Map Name Editing
--*Multiple Node Selection and Moving
--*Added color option for Window Selection
--*Fixed arrow behaviour of almost-linear links

--New In 0.2.1:
--*Fixed crash when linking texture with missing bitmap

--New In 0.1.19:
--*Highlight links and node on rollover
--*New color options
--*First steps to better Auto-Arrange mode (not wired yet)

--New In 0.1.18:
--*Support for all Material Node channels (except MorpherMat which is wicked)
--*Fixed bugs in Composite and Brick maps introduced in 0.1.17
--*Fixed pan speedup controls not working
--*Moved Input and Output links by one pixel

--New In 0.1.17:
--*Parent and Child Channel Display options
--*Partial Thumbnail display at left and right border
--*Arrows Size Option
--*Save state of Name,Type,Parent and Child buttons

--New In 0.1.16:
--*Electronic Scheme Link Option with Arrows
--*Changing Options updates view

--New In 0.1.15:
--*Basic Undo Option
--*Double-Rightclick returns to Move mode

--New In 0.1.14:
--*Circular Dependency Protection
--*Already Linked Maps display in Connect List

--New In 0.1.12:
--*Link / Unlink to Multiple Channels
--*Link as Copy option

--New In 0.1.11:
--*Unlink
--*Fix for map properties in Composite
--*Reload / Auto-Arrange 

--New In 0.1.10:
--*Struct Node storing
--*INI options storage
--*Thumbs Size and Spacing options
--*AppData position info storing

--ToDo:
--*Auto-Update Samples on external changes
--*Undo System
--*Node Selection

macroScript TreeMatoGraph category:"Bobo_s Tools" buttontext:"TMG"
(
global TreeMatoGraph_rollout

global TreeMatoGraph_Nodes 
global TreeMatoGraph_Links 


-- Size of Graphics Window, change to resize.
-- 800x600 should be the max., higher res. might be very slow!
-- 800x600 is about twice as slow as 600x400...
-- Window Resizing with the mouse coming in the next version!

local graph_width = 600
local graph_height = 400

if graph_width < 600 then graph_width = 600
if graph_width > 800 then graph_width = 800
if graph_height < 100 then graph_height = 100
if graph_height > 800 then graph_height = 800


local background_color = (color 150 150 170)
local thumb_width = 32
local thumb_height = 32
local thumb_hspacing = 15
local thumb_vspacing = 15
local show_lines_while_moving = false
local show_shaded_while_moving = false
local show_tooltip_on_rollover = false
local show_ghost_nodes = true
local quickdraw_skip_lines = 4
local link_parent_color = color 255 255 255
local link_child_color = color 0 0 100
local link_link_color = color 0 220 0
local link_unlink_color = color 200 0 50
local link_input_color = color 255 255 0
local link_output_color = color 220 0 0
local selection_window_color = color 255 0 0

local link_display_mode = 2
local link_arrow_size = 5

local map_sample = bitmap thumb_width thumb_height
local LButtonPressed = false
local MButtonPressed = false
local RButtonPressed = false
local current_node = 0
local clip_offset = Point2 0 0
local pan_offset = Point2 0 0
local origin_offset = Point2 -40 0
local start_pos = Point2 0 0
local already_used_array = #()
local already_used_index = #()
local collect_nodes = #()
local lights_array = #()
local lights_state = #()
local Thumb_label = #()
local Type_label = #()
local display_map_array = #()
local target_map
local source_map
local target_slots
local linked_slots
local type_counter = 0
local current_highlight = 0

local node_to_rename = undefined

local multiple_selection_count = 0
local multiple_selection_start 
local multiple_selection_end 
local multiple_selection_mode = false

local start_move_pos 

local old_node_positions = #()
local store_old_nodes = #()
local lonely_nodes = #()

global TreeMatoGraph_Currently_Edited_Tree 
local TreeMatoGraph_Undo_Buffer = undefined

local old_pos = [0,0]
local connect_source_node = 0
local ConnectMode = false
local DisconnectMode = false

struct tree_node (Name, XPos, YPos, Width, Height, ShadedColor, Thumbnail, MaxNode, Row, Column, Selected )

local storeIniFile
local restoreIniFile

	fn storeIniFile =
	(
		setIniFile = CreateFile ((GetDir #plugcfg)+"/TreeMatoGraph.ini")
		format "%\n" background_color to:setIniFile 
		format "%\n" thumb_width to:setIniFile 
		format "%\n" thumb_height to:setIniFile 
		format "%\n" thumb_hspacing to:setIniFile 
		format "%\n" thumb_vspacing to:setIniFile 
		format "%\n" show_lines_while_moving to:setIniFile 
		format "%\n" show_shaded_while_moving to:setIniFile 
		format "%\n" show_tooltip_on_rollover to:setIniFile 
		format "%\n" show_ghost_nodes to:setIniFile 
		format "%\n" quickdraw_skip_lines to:setIniFile 
		format "%\n" link_parent_color to:setIniFile 
		format "%\n" link_child_color to:setIniFile 
		format "%\n" link_display_mode to:setIniFile 
		format "%\n" link_arrow_size to:setIniFile 
		format "%\n" TreeMatoGraph_rollout.show_node_names.state to:setIniFile 
		format "%\n" TreeMatoGraph_rollout.show_node_class.state to:setIniFile 
		format "%\n" TreeMatoGraph_rollout.show_parent_channel.state to:setIniFile 
		format "%\n" TreeMatoGraph_rollout.show_child_channel.state to:setIniFile 
		format "%\n" link_link_color to:setIniFile 
		format "%\n" link_unlink_color to:setIniFile 
		format "%\n" link_output_color to:setIniFile 
		format "%\n" link_input_color to:setIniFile 
		format "%\n" selection_window_color to:setIniFile 
		close setIniFile 
	)	
	
	fn restoreIniFile =
	(
		getIniFile = OpenFile ((GetDir #plugcfg)+"/TreeMatoGraph.ini")
		if getIniFile != undefined then
		(
			try
			(
				background_color = execute ( readLine getIniFile )
				thumb_width = execute ( readLine getIniFile )
				thumb_height = execute ( readLine getIniFile )
				thumb_hspacing = execute ( readLine getIniFile )
				thumb_vspacing = execute ( readLine getIniFile )
				show_lines_while_moving = execute ( readLine getIniFile )
				show_shaded_while_moving = execute ( readLine getIniFile )
				show_tooltip_on_rollover = execute ( readLine getIniFile )
				show_ghost_nodes = execute ( readLine getIniFile )
				quickdraw_skip_lines = execute ( readLine getIniFile )
				link_parent_color = execute ( readLine getIniFile )
				link_child_color = execute ( readLine getIniFile )
				link_display_mode = execute ( readLine getIniFile ) 
				link_arrow_size = execute ( readLine getIniFile ) 
				TreeMatoGraph_rollout.show_node_names.state = execute ( readLine getIniFile ) 
				TreeMatoGraph_rollout.show_node_class.state = execute ( readLine getIniFile ) 
				TreeMatoGraph_rollout.show_parent_channel.state = execute ( readLine getIniFile ) 
				TreeMatoGraph_rollout.show_child_channel.state = execute ( readLine getIniFile ) 
				link_link_color = execute ( readLine getIniFile )
				link_unlink_color = execute ( readLine getIniFile )
				link_output_color = execute ( readLine getIniFile )
				link_input_color = execute ( readLine getIniFile )
				selection_window_color = execute ( readLine getIniFile )
			)catch()	
			close getIniFile 
		)	
	)
	
restoreIniFile()

local graph_bmp = bitmap graph_width graph_height color:background_color 
local background_plane = bitmap graph_width  graph_height color:background_color 
local back_plane = bitmap graph_width graph_height color:background_color 


	fn store_old_positions =
	(
		old_node_positions = #()
		store_old_nodes = #()
		lonely_nodes = #()
		for i = 1 to TreeMatoGraph_Nodes.count do
		(
			append store_old_nodes TreeMatoGraph_Nodes[i].MaxNode
			append old_node_positions (point2 TreeMatoGraph_Nodes[i].XPos TreeMatoGraph_Nodes[i].YPos)
			item_found = false
			for l in TreeMatoGraph_Links do
			(
				if l[1] == i or l[2] == i then
				(
					item_found = true
				)
			)
			if not item_found then
			(
				append lonely_nodes TreeMatoGraph_Nodes[i]
			)
		)
	)

	fn restore_old_positions =
	(
		for i in TreeMatoGraph_Nodes do
		(
			find_index = findItem store_old_nodes i.MaxNode
			if find_index > 0 then
			(
				i.XPos = old_node_positions[find_index].x
				i.YPos = old_node_positions[find_index].y
			)	
			for j = lonely_nodes.count to 1 by -1 do
			(
				if lonely_nodes[j].MaxNode == i.MaxNode then deleteItem lonely_nodes j
			)
		)
		for i in lonely_nodes do
		(
			append TreeMatoGraph_Nodes i
		)
	)

	fn restore_old_lonely_positions =
	(
		for i in TreeMatoGraph_Nodes do
		(
			for j = lonely_nodes.count to 1 by -1 do
			(
				if lonely_nodes[j].MaxNode == i.MaxNode then deleteItem lonely_nodes j
			)
		)
		for i in lonely_nodes do
		(
			append TreeMatoGraph_Nodes i
		)
	)
	
	
	--Check for Missing Maps
	
	local missingMaps = #()
	fn enumeratefn name = (append missingMaps name)
	fn checkForMissingMaps =
	(
		missingMaps = #()
	enumerateFiles enumeratefn #missing 
	missingMaps.count > 0 
	)	


rollout TreeMatoGraph_options_rollout "TreeMatoGraph Options"
(
	group "Thumbnails:"
	(
		spinner  options_thumb_width	"Width: "	range:[5,100,thumb_width] 		type:#integer fieldwidth:30 align:#left across:2
		spinner  options_spacing_width	"H Spacing:"	range:[1,100,thumb_hspacing] 	type:#integer fieldwidth:30 align:#right 
		spinner  options_thumb_height	"Height:"	    range:[5,100,thumb_height] 		type:#integer fieldwidth:30 align:#left across:2
		spinner  options_spacing_height	"V Spacing:"	range:[1,100,thumb_vspacing]	type:#integer fieldwidth:30 align:#right
	)
	
	group "Panning Performance:"
	(
		checkbox options_show_shaded 	"Single Color Nodes When Panning" 	checked:show_shaded_while_moving
		checkbox options_show_lines 	"Don't Draw Links When Panning" 	checked:(not show_lines_while_moving)
		checkbox options_ghost			"Ghost Nodes When Panning" 			checked:show_ghost_nodes 
		spinner  options_ghost_nth 		"Ghost Draw Every Nth Line:"		range:[2,10,quickdraw_skip_lines] type:#integer fieldwidth:30 offset:[0,1]
	)
	group "Display Options"
	(
--		checkbox options_tooltip 		"Tooltip on Rollover" 				checked:show_tooltip_on_rollover
		dropdownlist link_display		items:#("Diagonal Links","Straight Auto","Straight Horizontal","Straight Vertical") 	selection:(link_display_mode) width:120 across:2 align:#left offset:[-2,-2]
		spinner arrow_size "Arrow Size:" range:[3,15,link_arrow_size] fieldwidth:30 align:#right type:#integer
	)
	
	group "Colors"
	(
		colorpicker options_background_color "Background" 					color:background_color align:#right
		colorpicker options_parent_color 	"Link Color / Link At Parent" 	color:link_parent_color align:#right
		colorpicker options_child_color 	"Link At Child"  				color:link_child_color align:#right

		colorpicker options_output_color 	"Output Highlight Color" 		color:link_output_color align:#right
		colorpicker options_input_color 	"Input Highlight Color"  		color:link_input_color align:#right

		colorpicker options_link_color 		"Linking Color" 				color:link_link_color align:#right
		colorpicker options_unlink_color 	"Unlinking Color"  				color:link_unlink_color align:#right
		
		colorpicker options_select_color 	"Selection Window Color"  		color:selection_window_color align:#right
	)
	
	button cancel_options "Cancel" align:#left width:100 across:2
	button close_options "OK" align:#right width:100
	
	on options_thumb_width changed val do thumb_width = val
	on options_thumb_height changed val do thumb_height = val
	on options_spacing_width changed val do thumb_hspacing = val
	on options_spacing_height changed val do thumb_vspacing = val
	
	on options_show_lines changed state do show_lines_while_moving = not state
	on options_show_shaded changed state do show_shaded_while_moving = state
	on options_ghost changed state do show_ghost_nodes = state
	on options_ghost_nth changed val do quickdraw_skip_lines = val
--	on options_tooltip changed state do show_tooltip_on_rollover = state
	on link_display selected itm do link_display_mode = itm
	on arrow_size changed value do link_arrow_size = value
	
	
	
	on options_background_color changed state do 
	(
		background_color = state
		graph_bmp = bitmap graph_width graph_height color:background_color 
	)	
	on options_parent_color changed col do link_parent_color = col 
	on options_child_color changed col do link_child_color = col 
	
	on options_output_color changed col do link_output_color = col 
	on options_input_color changed col do link_input_color = col 
	
	
	on options_link_color changed col do link_link_color = col 
	on options_unlink_color changed col do link_unlink_color = col 

	on options_select_color changed col do selection_window_color = col 
	
	on cancel_options pressed do 
	(
		restoreIniFile()
		destroyDialog TreeMatoGraph_options_rollout
	)
	
	on close_options pressed do 
	(
		storeIniFile()
		destroyDialog TreeMatoGraph_options_rollout
		if TreeMatoGraph_Currently_Edited_Tree != undefined then 
		(
			graph_bmp = bitmap graph_width graph_height color:background_color 
			background_plane = bitmap graph_width  graph_height color:background_color 
			back_plane = bitmap graph_width graph_height color:background_color 
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog "Viewport Redraw Finished." 1
			TreeMatoGraph_rollout.displayNodeNames()
		)	
	)
)
	

rollout TreeMatoGraph_about_rollout "About TreeMatoGraph"
(
	label about02 "TreeMatoGraph"
	label about04 "v.0.2.5 - 08/02/2002"
	label about08 "Copyright (c) 2002 by Borislav Petrov"
	hyperlink email_hyperlink "bobo@email.archlab.tuwien.ac.at" address:"mailto:bobo@email.archlab.tuwien.ac.at" align:#center
	label about12 " "
	button close_about "Close" align:#center
	
	on close_about pressed do destroyDialog TreeMatoGraph_about_rollout
)

local TMG_RCMenu

RcMenu TMG_RCMenu
(
	subMenu "Tree"
	(
		menuItem tmg_new "New Tree..."
		menuItem tmg_clear "Clear Tree..."
		separator tmp_mat_sep1
		menuItem tmg_pick_selected "Get From Selected Object"
		menuItem tmg_pick_scene "Pick From Scene"
		menuItem tmg_pick_name  "Pick By Name"
		menuItem tmg_pick_medit "Get Current Material Editor Slot"
		separator tmp_mat_sep2
		menuItem tmg_update_current "Reload Current Tree"
		menuItem tmg_update_thumbs "Update Thumbnails"
		separator tmp_mat_sep3
		menuItem tmg_auto_arrange "Compact Node Arrangement"
	)


	subMenu "Edit"
	(
		menuItem tmg_undo "Undo"
--		menuItem tmg_redo "Redo"
		separator tmp_edit_sep1
		menuItem tmg_new_mat "New Material Node"
		menuItem tmg_new_map "New Map Node"
		separator tmp_edit_sep2
		menuItem tmg_options "Options..."
	)

	subMenu "Assign"
	(
		menuItem tmg_put_selection "Assign To Selection"
		menuItem tmg_put_medit "Assign To Current Material Editor Slot"
	)	
	
	subMenu "Display"
	(
		menuItem tmg_display_showmap  "Show Maps In Viewport"
		menuItem tmg_display_hidemap "Hide Maps In Viewport"
	)


	subMenu "Help"
	(
		menuItem tmg_help "TreeMatoGraph Webpage"
		separator tmg_help_sep1				
		menuItem tmg_boboland "More MAXScripts from Boboland"
		menuItem tmg_scriptspot "More MAXScripts from ScriptSpot"
		separator tmg_help_sep2		
		menuItem tmg_about "About TreeMatoGraph"
	)


	on tmg_undo picked do
	(
		if superclassof TreeMatoGraph_Undo_Buffer == Material then
		(
		
			temp_tree = TreeMatoGraph_Currently_Edited_Tree
--			store_old_positions()
			TreeMatoGraph_rollout.buildTree TreeMatoGraph_Undo_Buffer
			TreeMatoGraph_rollout.restoreTreePosition()
--			restore_old_positions()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog ("Undone Last Changes") 2
			TreeMatoGraph_Undo_Buffer = copy temp_tree
			TreeMatoGraph_Undo_Buffer.name = temp_tree.name
			TreeMatoGraph_rollout.displayNodeNames()
		)	
		else
		(
			TreeMatoGraph_rollout.UiLog ("Cannot Undo!") 4
		)
	)	
	
	on tmg_new_map picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree == undefined then
		(
			TreeMatoGraph_rollout.UiLog "Cannot Create Map - Tree Root Missing!" 4		
		)
		else
		(
			TreeMatoGraph_rollout.getNewMapNode [graph_width/2, graph_height/2]
			TreeMatoGraph_rollout.storeTreePosition()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1	
			TreeMatoGraph_rollout.displayNodeNames()
		)
	)
	
	on tmg_new_mat picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree == undefined then
		(
			mat = copy (materialBrowseDlg #mats)
			if mat != undefined then
			(
				TreeMatoGraph_rollout.CreateNewRoot mat 			
			)
			else
			(
				TreeMatoGraph_rollout.UiLog "Material Tree Root Creation Failed!" 4		
			)

		)
		else
		(
			TreeMatoGraph_rollout.getNewMatNode [graph_width/2, graph_height/2]
			TreeMatoGraph_rollout.storeTreePosition()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1	
			TreeMatoGraph_rollout.displayNodeNames()
		)	
	)
	
	on tmg_options picked do
	(
		CreateDialog TreeMatoGraph_options_rollout 240 475 modal:true
	)	

	on tmg_help picked do
	(
		shellLaunch "http://www.scriptspot.com/bobo/darkmoon/tmg/index.html" ""
	)	
	
	on tmg_update_thumbs picked do
	(
		TreeMatoGraph_rollout.updateAllThumbnails()
	)

	on tmg_boboland picked do
	(
		shellLaunch "http://www.scriptspot.com/bobo/" ""
	)		

	on tmg_scriptspot picked do
	(
		shellLaunch "http://www.scriptspot.com/" ""
	)		
	
	on tmg_about picked do
	(
		CreateDialog TreeMatoGraph_about_rollout 200 130 modal:true
	)
	
	
	fn filter_mats obj = obj.material != undefined	
	on tmg_pick_scene picked do
	(
		obj = pickobject filter:filter_mats
		if obj != undefined then
		(
			TreeMatoGraph_rollout.buildTree obj.material
			TreeMatoGraph_rollout.restoreTreePosition()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog ("Loaded Material From ["+ obj.name +"]") 2
		)
		else
		(
			TreeMatoGraph_rollout.UiLog "Nothing Selected." 4
		)
	)	
	
	on tmg_display_showmap picked do 
	(
		TreeMatoGraph_rollout.SMIV true
	)	
	on tmg_display_hidemap picked do 
	(
		TreeMatoGraph_rollout.SMIV false
	)	
	
	on tmg_pick_selected picked do
	(
		if selection.count == 1 then
		(
			If selection[1].material != undefined then 
			(
				TreeMatoGraph_rollout.buildTree selection[1].material
				TreeMatoGraph_rollout.restoreTreePosition()
				TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
				TreeMatoGraph_rollout.UiLog ("Loaded Material From ["+ selection[1].name +"]") 2
			)
			else
			(
				TreeMatoGraph_rollout.UiLog ("Object ["+ selection[1].name +"] Has No Material.") 4
			)		
		)
		else
		(
			TreeMatoGraph_rollout.UiLog ("Too Many Objects Selected.") 4
		)	
	)	


	on tmg_pick_name picked do
	(
		obj = selectByName single:true filter:filter_mats title:"TMG: Pick Object to Acquire Material..."
		if obj != undefined then
		(
			TreeMatoGraph_rollout.buildTree obj.material
			TreeMatoGraph_rollout.restoreTreePosition()				
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog ("Loaded Material From ["+ obj.name +"]") 2
			TreeMatoGraph_rollout.displayNodeNames()
		)
		else
		(
			TreeMatoGraph_rollout.UiLog "Nothing Selected." 4
		)		
	)	
	
	on tmg_update_current picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree != undefined then 
		(
			store_old_positions()
			TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree 
			TreeMatoGraph_rollout.restoreTreePosition()			
			restore_old_positions()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog "Reloaded Current Material Tree." 2
			TreeMatoGraph_rollout.displayNodeNames()
		)	
	)
	
	on tmg_auto_arrange  picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree != undefined then 
		(
			store_old_positions()
			TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree 
			restore_old_lonely_positions()
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog "Auto-Arranged Current Material Tree." 1
			TreeMatoGraph_rollout.displayNodeNames()
		)	
	)
	
	
	on tmg_pick_medit picked do
	(
		get_mat = getMeditMaterial (medit.getActiveMtlSlot())
		if superclassof get_mat == material then
		(
			try
			(
				TreeMatoGraph_Undo_Buffer = copy TreeMatoGraph_Currently_Edited_Tree 
				TreeMatoGraph_Undo_Buffer.name = TreeMatoGraph_Currently_Edited_Tree.name
			)catch()	
			TreeMatoGraph_Currently_Edited_Tree = get_mat
			TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree 
			TreeMatoGraph_rollout.restoreTreePosition()			
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
			TreeMatoGraph_rollout.UiLog ("Loaded Material ["+ TreeMatoGraph_Currently_Edited_Tree.name +"] From MEdit Slot "+ medit.getActiveMtlSlot() as string) 2
			TreeMatoGraph_rollout.displayNodeNames()
		)	
		else
		(
			TreeMatoGraph_rollout.UiLog "Cannot Get Bitmap as Root." 4
		)
	)
	
	on tmg_clear picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree == undefined then 
			q = true
		else
			q = querybox "About to CLEAR the TreeMatoGraph Tree?\nThis will remove all nodes.\nTo secure the current tree,\nASSIGN to an Object or Material Editor Slot!" title:"CLEAR TreeMatoGraph?"
		if q then
		(
			copy graph_bmp back_plane
			TreeMatoGraph_rollout.graph_bitmap.bitmap = back_plane
			TreeMatoGraph_Nodes = #()
			TreeMatoGraph_Links = #()	
			TreeMatoGraph_rollout.displayNodeNames()	
			try
			(
				TreeMatoGraph_Undo_Buffer = copy TreeMatoGraph_Currently_Edited_Tree 
				TreeMatoGraph_Undo_Buffer.name = TreeMatoGraph_Currently_Edited_Tree.name
			)catch()	
			TreeMatoGraph_Currently_Edited_Tree = undefined
			TreeMatoGraph_rollout.UiLog "TreeMatoGraph Cleared." 1
		)
		else
		(
			TreeMatoGraph_rollout.UiLog "TreeMatoGraph Clear Operation Canceled!" 4
		)
	)--end clear

	on tmg_new picked do
	(
		if TreeMatoGraph_Currently_Edited_Tree == undefined then 
			q = true
		else
			q = querybox "About to CREATE NEW TREE?\nThis will remove all current nodes.\nTo secure the current tree,\nASSIGN to an Object or Material Editor Slot!" title:"NEW TreeMatoGraph TREE?"
		if q then
		(
			mat = copy (materialBrowseDlg #mats)
			if mat != undefined then
			(
				TreeMatoGraph_rollout.CreateNewRoot mat 			
			)
			else
			(
				TreeMatoGraph_rollout.UiLog "Material Tree Root Creation Failed!" 4		
			)
		)
		else
		(
				TreeMatoGraph_rollout.UiLog "Material Tree Root Creation Canceled!" 4		
		)
	)--end new
	
	on tmg_put_selection picked do
	(
		selection.material = TreeMatoGraph_Currently_Edited_Tree
		TreeMatoGraph_rollout.UiLog ("Assigned To "+ selection.count as string+ " Selected Object(s)") 2
	)

	on tmg_put_medit picked do
	(
		medit_slot = medit.getActiveMtlSlot()
		setMeditMaterial medit_slot	TreeMatoGraph_Currently_Edited_Tree
		TreeMatoGraph_rollout.UiLog ("Assigned To MEdit Slot "+medit_slot as string) 2
	)
	
	
)--end RCMenu	



------------------
-- Link Rollout --
------------------
local TreeMatoGraph_link_map_rollout
local TreeMatoGraph_unlink_map_rollout


rollout TreeMatoGraph_link_map_rollout "TMG Link:"
(
	button link_child  "Link Child" width:90 across:2 align:#left
	checkbutton as_copy "As Copy" checked:false width:80 align:#right

	edittext label_source  ""
	label label_target0 "To Parent"
	edittext label_target  ""
	multilistbox map_list height:20
	
	
	
	fn link_function itm =
	(
		try
		(
			TreeMatoGraph_Undo_Buffer = copy TreeMatoGraph_Currently_Edited_Tree 
			TreeMatoGraph_Undo_Buffer.name = TreeMatoGraph_Currently_Edited_Tree.name
		)catch()	
		store_old_positions()
		local original_map = undefined
		if target_slots[itm][3] == "Material" and superclassof source_map != Material then 
		(
				TreeMatoGraph_rollout.UiLog ("Cannot Link: Material Expected.") 4
				return false
		)
		else
		(
			if target_slots[itm][3] == "Map" and superclassof source_map != TextureMap then 
			(
				TreeMatoGraph_rollout.UiLog ("Cannot Link: Map Exprected.") 4
				return false
			)
			else
			(
				if target_slots[itm][2] == "Property" then 
				(
					original_map = getProperty target_map target_slots[itm][1]
					if superclassof original_map == TextureMap or superclassof original_map == Material then
					(
						for i in TreeMatoGraph_Nodes do
						(
							if i.MaxNode == original_map and findItem lonely_nodes i == 0 then
							(
								append lonely_nodes i
							)	
						)	
					)
				if as_copy.checked then 	
					setProperty target_map target_slots[itm][1] (copy source_map)
				else	
					setProperty target_map target_slots[itm][1] source_map
				)
				else	
				(
					try(original_map = execute ("target_map"+target_slots[itm][2]+".object"))catch()
					if superclassof original_map == TextureMap or superclassof original_map == Material then
					(
						for i in TreeMatoGraph_Nodes do
						(
							if i.MaxNode == original_map and findItem lonely_nodes i == 0 then
							(
								append lonely_nodes i
							)	
						)	
					)
					txt = "target_map"+target_slots[itm][2]+ "= source_map\n"
					try(execute ( txt ))catch()
				)
			)
		)	
		true	
	)
	
	fn updateTreeAfterLink success=
	(
		local selected_channels = (map_list.selection as array).count
		destroyDialog TreeMatoGraph_link_map_rollout
		if success then
		(
			TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree
			restore_old_positions()		
			TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1	
			TreeMatoGraph_rollout.displayNodeNames()
			if selected_channels  == 1 then
			(
				for i in map_list.selection do
					TreeMatoGraph_rollout.UiLog ("Linked "+ map_list.items[i] +": " +source_map.name + " to " +target_map.name ) 2		
			)
			else	
			(
				TreeMatoGraph_rollout.UiLog ("Linked "+ selected_channels as string +" Channels: " +source_map.name + " to " +target_map.name ) 2		
			)
		)	
	)
	
	on map_list doubleclicked itm do 
	(
		success = link_function itm
		updateTreeAfterLink success
	)	
	on link_child pressed do 
	(
		success_counter = 0
		for i in map_list.selection do
		(
			success = link_function i
			if success then success_counter += 1
		)	
		updateTreeAfterLink (success_counter > 0)
	)
)

--------------------
-- Unlink Rollout --
--------------------

rollout TreeMatoGraph_unlink_map_rollout "TMG Unlink:"
(
	button unlink_child  "Unlink Child" width:180
	edittext label_source  ""
	label label_target0 "From Parent"
	edittext label_target  ""
	multilistbox map_list height:20
	
	fn unlink_function itm =
	(
		try
		(
			TreeMatoGraph_Undo_Buffer = copy TreeMatoGraph_Currently_Edited_Tree 
			TreeMatoGraph_Undo_Buffer.name = TreeMatoGraph_Currently_Edited_Tree.name
		)catch()	
	
		store_old_positions()
		local original_map = undefined
		if linked_slots[itm][2] == "Property" then 
		(
			original_map = getProperty target_map linked_slots[itm][1]
			if superclassof original_map == TextureMap or superclassof original_map == Material then
			(
				for i in TreeMatoGraph_Nodes do
				(
					if i.MaxNode == original_map and findItem lonely_nodes i == 0 then
					(
						append lonely_nodes i
					)	
				)	
			)
			try
			(
				setProperty target_map linked_slots[itm][1] undefined
			)
			catch
			(
			)
		)
		else	
		(
			try(original_map = execute ("target_map"+linked_slots[itm][2]+".object"))catch()
			if superclassof original_map == TextureMap or superclassof original_map == Material then
			(
				for i in TreeMatoGraph_Nodes do
				(
					if i.MaxNode == original_map and findItem lonely_nodes i == 0 then
					(
						append lonely_nodes i
					)	
				)	
			)
			txt = "target_map"+linked_slots[itm][2]+ "= undefined\n"
			try(execute ( txt ))catch()
		)
	)
	
	fn updateTreeAfterUnlink =
	(
		local selected_channels = (map_list.selection as array).count
		destroyDialog TreeMatoGraph_unlink_map_rollout
		TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree
		restore_old_positions()		
		TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1	
		TreeMatoGraph_rollout.displayNodeNames()
		if selected_channels  == 1 then
		(
			for i in map_list.selection do
				TreeMatoGraph_rollout.UiLog ("Unlinked "+ map_list.items[i] +": " +source_map.name + " from " +target_map.name ) 2		
		)
		else	
		(
			TreeMatoGraph_rollout.UiLog ("Unlinked "+ selected_channels as string +" Channels: " +source_map.name + " from " +target_map.name ) 2		
		)
	)
	
	on map_list doubleclicked itm do 
	(
		unlink_function itm 
		updateTreeAfterUnlink()
	)	
	on unlink_child pressed do 
	(
		for i in map_list.selection do
		(
			unlink_function i
		)
		updateTreeAfterUnlink()
	)	
)


------------------
-- Main Rollout --
------------------

local col_array = #()

rollout TreeMatoGraph_rollout  "TreeMatoGraph v0.2.5"
(
	bitmap graph_bitmap bitmap:graph_bmp width:graph_width height:graph_height pos:[0,0]
	checkbutton status_bar1 "" checked:true align:#left pos:[250,422] width:335 highlightcolor:(color 200 200 255) 
	checkbutton status_bar2 "" checked:true align:#left pos:[250,422] width:335 highlightcolor:(color 200 255 200) 
	checkbutton status_bar4 "" checked:true align:#left pos:[250,422] width:335 highlightcolor:(color 255 200 200) 
	checkbutton status_bar6 "" checked:true align:#left pos:[250,422] width:335 highlightcolor:(color 255 255 200) 
	fn defineColArray =
	(
		col_array = #( status_bar1, status_bar2, status_bar2, status_bar4, status_bar4, status_bar6 )
	)
	
	fn UiLog txt col_index =
	(
		for i = 1 to col_array.count where i != col_index do 
		(
			col_array[i].pos = [1000,1000]
		)	
		(col_array[col_index]).text = txt
		(col_array[col_index]).pos =[265,graph_height+22]
	)	
	
	on status_bar1 changed state do status_bar1.checked = true
	on status_bar2 changed state do status_bar2.checked = true
	on status_bar4 changed state do status_bar4.checked = true
	on status_bar6 changed state do status_bar6.checked = true
	
	checkbutton move_mode "MOVE" pos:[5,graph_height+1] width:45 height:20 checked:true
	checkbutton del_mode "DEL" pos:[5,graph_height+22] width:45 height:20 checked:false enabled:false
	
	checkbutton link_mode "LINK" pos:[50,graph_height+1] width:45 height:20  checked:false
	checkbutton unlink_mode "UNLNK" pos:[50,graph_height+22] width:45 height:20  checked:false

	button rescan_tree "RELOAD" pos:[100,graph_height+1] width:55 height:41
	
	checkbutton show_node_names "Name" checked:true  pos:[160,graph_height+1] width:50 height:20
	checkbutton show_node_class "Type" checked:false pos:[210,graph_height+1] width:50 height:20
	checkbutton show_parent_channel "Outputs" checked:true pos:[160,graph_height+22] width:50 height:20 highlightcolor:(link_output_color) tooltip:"Show Channels in Parents the Current Node is Linked to."
	checkbutton show_child_channel "Inputs" checked:true pos:[210,graph_height+22] width:50 height:20 highlightcolor:(link_input_color) tooltip:"Show Channels that Children of the Current Node are Linked to."


--	mapbutton map_button "Map" pos:[265,401] width:120 height:20 
	edittext map_edit_name width:200 pos:[261,graph_height+2]
	
	
	on link_mode changed state do
	(
	    if state then
			move_mode.checked = del_mode.checked = unlink_mode.checked = false
		else
			link_mode.checked = true
	)	
	on move_mode changed state do 
	(
	    if state then
			del_mode.checked = unlink_mode.checked = link_mode.checked = false
		else	
			move_mode.checked = true
	)	
	on del_mode changed state do 
	(
		if state then
			move_mode.checked = unlink_mode.checked = link_mode.checked = false
		else	
			del_mode.checked = true
	)	
	on unlink_mode changed state do 
	(
		if state then
			move_mode.checked = del_mode.checked = link_mode.checked = false
		else
			unlink_mode.checked = true
	)	
	
	on show_node_names changed state do storeIniFile()
	on show_node_class changed state do storeIniFile()
	on show_parent_channel changed state do storeIniFile()
	on show_child_channel changed state do storeIniFile()
	
	label tooltip_label "" pos:[1000,1000]
	
	label thumb_label_01 "" pos:[1000,1000] align:#left
	label thumb_label_02 "" pos:[1000,1000] align:#left
	label thumb_label_03 "" pos:[1000,1000] align:#left
	label thumb_label_04 "" pos:[1000,1000] align:#left
	label thumb_label_05 "" pos:[1000,1000] align:#left
	label thumb_label_06 "" pos:[1000,1000] align:#left
	label thumb_label_07 "" pos:[1000,1000] align:#left
	label thumb_label_08 "" pos:[1000,1000] align:#left
	label thumb_label_09 "" pos:[1000,1000] align:#left
	label thumb_label_10 "" pos:[1000,1000] align:#left

	label thumb_label_11 "" pos:[1000,1000] align:#left
	label thumb_label_12 "" pos:[1000,1000] align:#left
	label thumb_label_13 "" pos:[1000,1000] align:#left
	label thumb_label_14 "" pos:[1000,1000] align:#left
	label thumb_label_15 "" pos:[1000,1000] align:#left
	label thumb_label_16 "" pos:[1000,1000] align:#left
	label thumb_label_17 "" pos:[1000,1000] align:#left
	label thumb_label_18 "" pos:[1000,1000] align:#left
	label thumb_label_19 "" pos:[1000,1000] align:#left
	label thumb_label_20 "" pos:[1000,1000] align:#left

	label thumb_label_21 "" pos:[1000,1000] align:#left
	label thumb_label_22 "" pos:[1000,1000] align:#left
	label thumb_label_23 "" pos:[1000,1000] align:#left
	label thumb_label_24 "" pos:[1000,1000] align:#left
	label thumb_label_25 "" pos:[1000,1000] align:#left
	label thumb_label_26 "" pos:[1000,1000] align:#left
	label thumb_label_27 "" pos:[1000,1000] align:#left
	label thumb_label_28 "" pos:[1000,1000] align:#left
	label thumb_label_29 "" pos:[1000,1000] align:#left
	label thumb_label_30 "" pos:[1000,1000] align:#left
	
	label thumb_label_31 "" pos:[1000,1000] align:#left
	label thumb_label_32 "" pos:[1000,1000] align:#left
	label thumb_label_33 "" pos:[1000,1000] align:#left
	label thumb_label_34 "" pos:[1000,1000] align:#left
	label thumb_label_35 "" pos:[1000,1000] align:#left
	label thumb_label_36 "" pos:[1000,1000] align:#left
	label thumb_label_37 "" pos:[1000,1000] align:#left
	label thumb_label_38 "" pos:[1000,1000] align:#left
	label thumb_label_39 "" pos:[1000,1000] align:#left
	label thumb_label_40 "" pos:[1000,1000] align:#left	

	fn DefineThumbLabel = 
	(
		Thumb_label = #(thumb_label_01, thumb_label_02, thumb_label_03, thumb_label_04, thumb_label_05, thumb_label_06, thumb_label_07, thumb_label_08, thumb_label_09, thumb_label_10,
						thumb_label_11, thumb_label_12, thumb_label_13, thumb_label_14, thumb_label_15, thumb_label_16, thumb_label_17, thumb_label_18, thumb_label_19, thumb_label_20,
						thumb_label_21, thumb_label_22, thumb_label_23, thumb_label_24, thumb_label_25, thumb_label_26, thumb_label_27, thumb_label_28, thumb_label_29, thumb_label_30,
						thumb_label_31, thumb_label_32, thumb_label_33, thumb_label_34, thumb_label_35, thumb_label_36, thumb_label_37, thumb_label_38, thumb_label_39, thumb_label_40
		)
	)
	
	
	label type_label_01 "" pos:[1000,1000] align:#left
	label type_label_02 "" pos:[1000,1000] align:#left
	label type_label_03 "" pos:[1000,1000] align:#left
	label type_label_04 "" pos:[1000,1000] align:#left
	label type_label_05 "" pos:[1000,1000] align:#left
	label type_label_06 "" pos:[1000,1000] align:#left
	label type_label_07 "" pos:[1000,1000] align:#left
	label type_label_08 "" pos:[1000,1000] align:#left
	label type_label_09 "" pos:[1000,1000] align:#left
	label type_label_10 "" pos:[1000,1000] align:#left

	label type_label_11 "" pos:[1000,1000] align:#left
	label type_label_12 "" pos:[1000,1000] align:#left
	label type_label_13 "" pos:[1000,1000] align:#left
	label type_label_14 "" pos:[1000,1000] align:#left
	label type_label_15 "" pos:[1000,1000] align:#left
	label type_label_16 "" pos:[1000,1000] align:#left
	label type_label_17 "" pos:[1000,1000] align:#left
	label type_label_18 "" pos:[1000,1000] align:#left
	label type_label_19 "" pos:[1000,1000] align:#left
	label type_label_20 "" pos:[1000,1000] align:#left

	label type_label_21 "" pos:[1000,1000] align:#left
	label type_label_22 "" pos:[1000,1000] align:#left
	label type_label_23 "" pos:[1000,1000] align:#left
	label type_label_24 "" pos:[1000,1000] align:#left
	label type_label_25 "" pos:[1000,1000] align:#left
	label type_label_26 "" pos:[1000,1000] align:#left
	label type_label_27 "" pos:[1000,1000] align:#left
	label type_label_28 "" pos:[1000,1000] align:#left
	label type_label_29 "" pos:[1000,1000] align:#left
	label type_label_30 "" pos:[1000,1000] align:#left
	
	label type_label_31 "" pos:[1000,1000] align:#left
	label type_label_32 "" pos:[1000,1000] align:#left
	label type_label_33 "" pos:[1000,1000] align:#left
	label type_label_34 "" pos:[1000,1000] align:#left
	label type_label_35 "" pos:[1000,1000] align:#left
	label type_label_36 "" pos:[1000,1000] align:#left
	label type_label_37 "" pos:[1000,1000] align:#left
	label type_label_38 "" pos:[1000,1000] align:#left
	label type_label_39 "" pos:[1000,1000] align:#left
	label type_label_40 "" pos:[1000,1000] align:#left	

	fn DefineTypeLabel = 
	(
		Type_label = #(type_label_01, type_label_02, type_label_03, type_label_04, type_label_05, type_label_06, type_label_07, type_label_08, type_label_09, type_label_10,
						type_label_11, type_label_12, type_label_13, type_label_14, type_label_15, type_label_16, type_label_17, type_label_18, type_label_19, type_label_20,
						type_label_21, type_label_22, type_label_23, type_label_24, type_label_25, type_label_26, type_label_27, type_label_28, type_label_29, type_label_30,
						type_label_31, type_label_32, type_label_33, type_label_34, type_label_35, type_label_36, type_label_37, type_label_38, type_label_39, type_label_40
		)
	)
	
	
	fn displayNodeNames =
	(
		for n = 1 to TreeMatoGraph_Nodes.count do
		(
			Thumb_label[n].pos = [ TreeMatoGraph_Nodes[n].XPos+origin_offset.x, TreeMatoGraph_Nodes[n].YPos+origin_offset.y ] + [0,TreeMatoGraph_Nodes[n].Height]
			map_name = ""
			if show_node_names.checked then map_name += TreeMatoGraph_Nodes[n].MaxNode.name + " "
			if show_node_class.checked then map_name += "["+ (classof TreeMatoGraph_Nodes[n].MaxNode as string)+"]"
			Thumb_label[n].text = map_name
		)
		for n = TreeMatoGraph_Nodes.count+1 to Thumb_label.count do
		(
			Thumb_label[n].pos = [1000,1000]
		)
	)	
	
	fn clearNodeTypes = 
	(
		for n in Type_label do n.pos = [1000,1000]
	)

	fn showConnectionType from_node to_pos node_label =
	(
		from_middle_x = TreeMatoGraph_Nodes[from_node].XPos+TreeMatoGraph_Nodes[from_node].Width/2.0
		from_middle_y = TreeMatoGraph_Nodes[from_node].YPos+TreeMatoGraph_Nodes[from_node].Height/2.0
		
		to_middle_x = TreeMatoGraph_Nodes[to_node].XPos+TreeMatoGraph_Nodes[to_node].Width/2.0
		to_middle_y = TreeMatoGraph_Nodes[to_node].YPos+TreeMatoGraph_Nodes[to_node].Height/2.0
		
		middle_x = ( from_middle_x + to_middle_x ) / 2.0
		middle_y = ( from_middle_y + to_middle_y ) / 2.0
		node_label.pos = Point2 middle_x middle_y
	)	
	
	
	fn displayConnectionTypes =
	(
		for n = 1 to TreeMatoGraph_Nodes.count do
		(
			showConnectionType n
			Thumb_label[n].pos = [ TreeMatoGraph_Nodes[n].XPos+origin_offset.x, TreeMatoGraph_Nodes[n].YPos+origin_offset.y ] + [0,TreeMatoGraph_Nodes[n].Height]
			map_name = ""
			if show_node_names.checked then map_name += TreeMatoGraph_Nodes[n].MaxNode.name + " "
			if show_node_class.checked then map_name += "["+ (classof TreeMatoGraph_Nodes[n].MaxNode as string)+"]"
			Thumb_label[n].text = map_name
		)
		for n = TreeMatoGraph_Nodes.count+1 to Thumb_label.count do
		(
			Thumb_label[n].pos = [1000,1000]
		)
	)	
	

	fn drawNode node_index b_plane full_draw =
	(
		draw_image = TreeMatoGraph_Nodes[node_index].thumbnail != undefined
		if show_shaded_while_moving and full_draw == 0 then draw_image = false
		if full_draw == 0 and show_ghost_nodes then quickdraw_bitmaps_skip_lines = quickdraw_skip_lines else quickdraw_bitmaps_skip_lines = 1
		if TreeMatoGraph_Nodes[node_index].Selected then quickdraw_bitmaps_skip_lines = 1.5
		
		if avguard_dlx_ver != undefined and not TreeMatoGraph_Nodes[node_index].Selected then
		(
			DrawIt = true
			actual_origin_x = 0
			actual_origin_y = 0
			actual_target_x = TreeMatoGraph_Nodes[node_index].XPos+origin_offset.x
			actual_target_y = TreeMatoGraph_Nodes[node_index].YPos+origin_offset.y
			actual_width = TreeMatoGraph_Nodes[node_index].Width
			actual_height = TreeMatoGraph_Nodes[node_index].Height
			if actual_target_x < 0 then 
			(
				actual_width += actual_target_x
				actual_origin_x -= actual_target_x
				actual_target_x = 0
				if abs(actual_width) > actual_width then DrawIt = false
			)	
			if actual_target_y < 0 then 
			(
				actual_height += actual_target_y
				actual_origin_y -= actual_target_y
				actual_target_y = 0
				if abs(actual_height) > actual_height then DrawIt = false
			)	
			
			if actual_target_x >= graph_width-1 then DrawIt = false
			if actual_target_y >= graph_height-1 then DrawIt = false
			
			if actual_width < 0 then actual_width = 0
			if DrawIt then
				pasteBitmap TreeMatoGraph_Nodes[node_index].thumbnail b_plane (Box2 actual_origin_x actual_origin_y actual_width actual_height) [actual_target_x, actual_target_y]
		)
		else
		(
			target_x = TreeMatoGraph_Nodes[node_index].XPos+origin_offset.x
			get_pixels_amount = TreeMatoGraph_Nodes[node_index].Width
			get_pixles_offset = 0
			if target_x < 1 and (abs target_x) < get_pixels_amount then 
			(
				get_pixles_offset = abs(target_x) 
				get_pixels_amount -= get_pixles_offset 
			)	
			if target_x+get_pixels_amount > graph_width then get_pixels_amount -= target_x+get_pixels_amount-graph_width
		
			if draw_image then
			(
				for j = 1 to TreeMatoGraph_Nodes[node_index].Height by quickdraw_bitmaps_skip_lines do
				(
					pixel_array = getPixels TreeMatoGraph_Nodes[node_index].thumbnail [get_pixles_offset,j-1] get_pixels_amount 
					setPixels b_plane [target_x+get_pixles_offset, TreeMatoGraph_Nodes[node_index].YPos+j-1+origin_offset.y] pixel_array
				)	
			)
			else
			(
				pixel_array = #()
				for i = 1 to get_pixels_amount do append pixel_array TreeMatoGraph_Nodes[node_index].ShadedColor
				for j = 1 to TreeMatoGraph_Nodes[node_index].Height by quickdraw_bitmaps_skip_lines do
				(
					setPixels b_plane [target_x+get_pixles_offset , TreeMatoGraph_Nodes[node_index].YPos+j-1+origin_offset.y] pixel_array
				)	
			)
		)
		b_plane
	)
	
	fn drawArrow b_plane pos direction size color1 =
	(
		pixel_array = #()
		case direction of
		(
			1: (
 				append pixel_array color1
				for i = 1 to size do
				(
					append pixel_array color1
					append pixel_array color1
					setPixels b_plane [pos.x-i,pos.y+i-size/2.0] pixel_array
				)
			)	
			2: (
				for i = 0 to size do
				(
					append pixel_array color1
					setPixels b_plane [pos.x-size+size/2.0,pos.y-size+i] pixel_array
					setPixels b_plane [pos.x-size+size/2.0,pos.y+size-i] pixel_array
				)
			)	
			3: (
				append pixel_array color1
				for i = 1 to size do
				(
					append pixel_array color1
					append pixel_array color1
					setPixels b_plane [pos.x-i,pos.y-i+size/2.0] pixel_array
				)
			)	
			4: (
				for i = 0 to size do
				(
					append pixel_array color1
					setPixels b_plane [pos.x+size-i-size/2.0,pos.y-size+i] pixel_array
					setPixels b_plane [pos.x+size-i-size/2.0,pos.y+size-i] pixel_array
				)
			)			
			
		)	
	b_plane	
	)
	
	
	fn drawBorder node color1 b_plane =
	(
		
		pos_x = TreeMatoGraph_Nodes[node].XPos
		pos_y = TreeMatoGraph_Nodes[node].YPos
		width = TreeMatoGraph_Nodes[node].Width
		height = TreeMatoGraph_Nodes[node].Height
		pixel_array = #()
		for i = 1 to width do
		(
			append pixel_array color1
		)
		setpixels b_plane [pos_x+origin_offset.x, pos_y+origin_offset.y] pixel_array
		setpixels b_plane [pos_x+origin_offset.x, pos_y+origin_offset.y+height-1] pixel_array
		pixel_array = #(color1)
		for i = 1 to height-1 do
		(
			setpixels b_plane [pos_x+origin_offset.x, pos_y+origin_offset.y+i] pixel_array
			setpixels b_plane [pos_x+origin_offset.x+width-1, pos_y+origin_offset.y+i] pixel_array
		)
	b_plane
	)
	
	
	fn drawSelectionWindow b_plane from_pos to_pos =
	(
		pixel_array = #()
		if from_pos.x > to_pos.x then
		(
			upper_left_x = to_pos.x
			lower_right_x = from_pos.x
		)	
		else
		(
			upper_left_x = from_pos.x
			lower_right_x = to_pos.x
		)
		if from_pos.y > to_pos.y then
		(
			upper_left_y = to_pos.y
			lower_right_y = from_pos.y
		)	
		else
		(
			upper_left_y = from_pos.y
			lower_right_y = to_pos.y
		)	
		w_width = lower_right_x-upper_left_x
		w_height = lower_right_y-upper_left_y
		for i = 1 to w_width do append pixel_array selection_window_color
		setPixels b_plane [upper_left_x, upper_left_y] pixel_array 
		setPixels b_plane [upper_left_x, lower_right_y] pixel_array 
		pixel_array = #(selection_window_color)
		for i = 1 to w_height do
		(
			setPixels b_plane [upper_left_x, upper_left_y+i] pixel_array 
			setPixels b_plane [lower_right_x, upper_left_y+i] pixel_array 
		)
	b_plane 
	)
	

	
	fn drawConnection from_point to_point b_plane dash_step dash_length color1 color2 mode=
	(
		start_color = color1 
		end_color = color2   

		from_point.x += origin_offset.x + 1
		from_point.y += origin_offset.y + 1
		to_point.x += origin_offset.x - 1
		to_point.y += origin_offset.y - 1
		
		if from_point.x < 0 then from_point.x = 0
		if to_point.x < 0 then to_point.x = 0

		if from_point.x > graph_width then from_point.x = graph_width
		if to_point.x > graph_width then to_point.x = graph_width
		
		mid_point = (from_point + to_point)/2.0
		sub_point = (from_point - to_point)
		abs_sub_point = Point2 (abs sub_point.x) (abs sub_point.y)
		case mode of
		(
			1: horizontal_first = abs_sub_point.x > abs_sub_point.y 
			2: horizontal_first = true
			3: horizontal_first = false
		)	
		
		if horizontal_first then
		(
			pixel_array = #()
			for x = 1 to abs_sub_point.x/2 do
			(
				append pixel_array start_color
			)
			if sub_point.x < 0 then
			(
				setPixels b_plane [from_point.x,from_point.y] pixel_array
				setPixels b_plane [mid_point.x,to_point.y] pixel_array
				
			)
			else
			(
				setPixels b_plane [to_point.x,to_point.y] pixel_array
				setPixels b_plane [mid_point.x,from_point.y] pixel_array
			)	
			pixel_array = #(start_color)
			if sub_point.y < 0 then
			(
				for y =  1 to abs_sub_point.y do
					setPixels b_plane [mid_point.x,from_point.y+y] pixel_array
				if abs_sub_point.y < link_arrow_size then
				(
					if sub_point.x < 0 then
						drawArrow b_plane mid_point 4 link_arrow_size start_color
					else	
						drawArrow b_plane mid_point 2 link_arrow_size start_color
				)
				else	
					drawArrow b_plane mid_point 1 link_arrow_size start_color
			)	
			else
			(
				for y =  1 to abs_sub_point.y do
					setPixels b_plane [mid_point.x,from_point.y-y] pixel_array
				if abs_sub_point.y < link_arrow_size then
				(
					if sub_point.x < 0 then
						drawArrow b_plane mid_point 4 link_arrow_size start_color
					else	
						drawArrow b_plane mid_point 2 link_arrow_size start_color
				)		
				else	
					drawArrow b_plane mid_point 3 link_arrow_size start_color
			)
		)
		else
		(
			pixel_array = #()
			for x = 1 to abs_sub_point.x do
			(
				append pixel_array start_color
			)
			if sub_point.x < 0 then
			(
				setPixels b_plane [from_point.x,mid_point.y] pixel_array
				if abs_sub_point.x < link_arrow_size then
				(
					if sub_point.y < 0 then
						drawArrow b_plane mid_point 1 link_arrow_size start_color
					else	
						drawArrow b_plane mid_point 3 link_arrow_size start_color
				)		
				else	
					drawArrow b_plane mid_point 4 link_arrow_size start_color
			)
			else
			(
				setPixels b_plane [to_point.x,mid_point.y] pixel_array
				if abs_sub_point.x < link_arrow_size then
				(
					if sub_point.y < 0 then
						drawArrow b_plane mid_point 1 link_arrow_size start_color
					else	
						drawArrow b_plane mid_point 3 link_arrow_size start_color
				)	
				else	
					drawArrow b_plane mid_point 2 link_arrow_size start_color
			)	
			pixel_array = #(start_color)
			if sub_point.y < 0 then
			(
				for y = 1 to abs_sub_point.y/2 do
				(
					setPixels b_plane [from_point.x,from_point.y+y] pixel_array
					setPixels b_plane [to_point.x,to_point.y-y] pixel_array
				)	
				
			)	
			else
			(
				for y =  1 to abs_sub_point.y/2 do
				(
					setPixels b_plane [from_point.x,from_point.y-y] pixel_array
					setPixels b_plane [to_point.x,to_point.y+y] pixel_array
				)	
			)
		)	
	b_plane	
	)	
	
	
	
	fn drawDiagonalConnection from_point to_point b_plane dash_step dash_length color1 color2 =
	(
		start_color = color1 --link_parent_color
		end_color = color2   --link_child_color
		dx = to_point.x - from_point.x
		dy = to_point.y - from_point.y
		if abs(dx) > abs(dy) then steps = abs(dx) else steps = abs(dy)
		linex = from_point.x+origin_offset.x
		liney = from_point.y+origin_offset.y
		stepdx = dx/ steps as float
		stepdy = dy/ steps as float
		color_delta = (end_color-start_color)/steps as float
		current_color = start_color
		dash_count = 0
		for i = 1 to steps do
		(
			linex += stepdx
			liney += stepdy
			current_color += color_delta
			pixel_array = #(current_color,current_color)
			dash_count += 1
			if dash_count <=dash_length then
				setPixels b_plane [linex,liney] pixel_array
			
			if dash_count == dash_step then
				dash_count = 0
		)
	b_plane	
	)

	fn drawMoveNodeLines b_plane except dash_step dash_length color1 color2=
	(
		for l in TreeMatoGraph_Links do
		(
			if l[1] > 0 and l[2] > 0 then
			(
				if l[1] == except or l[2] == except then
				(
					from_middle_x = TreeMatoGraph_Nodes[l[1]].XPos+TreeMatoGraph_Nodes[l[1]].Width/2.0
					from_middle_y = TreeMatoGraph_Nodes[l[1]].YPos+TreeMatoGraph_Nodes[l[1]].Height/2.0
	
					to_middle_x = TreeMatoGraph_Nodes[l[2]].XPos+TreeMatoGraph_Nodes[l[2]].Width/2.0
					to_middle_y = TreeMatoGraph_Nodes[l[2]].YPos+TreeMatoGraph_Nodes[l[2]].Height/2.0
					case link_display_mode of
					(
						1: (
							if l[2] == except then 
								b_plane = drawDiagonalConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_output_color link_output_color
							else
								b_plane = drawDiagonalConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_input_color link_input_color
							)
						2: (
							if l[2] == except then 
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_output_color link_output_color 1
							else
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_input_color link_input_color 1
						)	
						3: (
							if l[2] == except then 
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_output_color link_output_color 2
							else
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_input_color link_input_color 2
						)	
						4: (
							if l[2] == except then 
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_output_color link_output_color 3
							else
								b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_input_color link_input_color 3
						)	
						
					)	
				)		
			)		
		)
	b_plane	
	)
	

	fn drawConnectNodeLine b_plane from_node to_pos dash_step dash_length color1 color2 =
	(
		from_middle_x = TreeMatoGraph_Nodes[from_node].XPos+TreeMatoGraph_Nodes[from_node].Width/2.0
		from_middle_y = TreeMatoGraph_Nodes[from_node].YPos+TreeMatoGraph_Nodes[from_node].Height/2.0
		case link_display_mode of
		(
			1: b_plane = drawDiagonalConnection (Point2 from_middle_x from_middle_y) to_pos b_plane dash_step dash_length color1 color2
			2: b_plane = drawConnection to_pos (Point2 from_middle_x from_middle_y)  b_plane dash_step dash_length color1 color2 1
			3: b_plane = drawConnection to_pos (Point2 from_middle_x from_middle_y)  b_plane dash_step dash_length color1 color2 2
			4: b_plane = drawConnection to_pos (Point2 from_middle_x from_middle_y)  b_plane dash_step dash_length color1 color2 3
		)			
	b_plane	
	)
	
	

	fn drawAllNodes b_plane except dash_step dash_length full_draw=
	(
		if full_draw == 1 or show_lines_while_moving then
		(
		for l in TreeMatoGraph_Links do
		(
			if l[1] > 0 and l[2] > 0 then
			(
				if l[1] != except and l[2] != except then
				(
					from_middle_x = TreeMatoGraph_Nodes[l[1]].XPos+TreeMatoGraph_Nodes[l[1]].Width/2.0
					from_middle_y = TreeMatoGraph_Nodes[l[1]].YPos+TreeMatoGraph_Nodes[l[1]].Height/2.0
	
					to_middle_x = TreeMatoGraph_Nodes[l[2]].XPos+TreeMatoGraph_Nodes[l[2]].Width/2.0
					to_middle_y = TreeMatoGraph_Nodes[l[2]].YPos+TreeMatoGraph_Nodes[l[2]].Height/2.0
				
					if current_node == l[1] or current_node == l[2] then
					(
						case link_display_mode of
						(
							1: b_plane = drawDiagonalConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_parent_color link_child_color
							2: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_parent_color link_child_color 1
							3: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_parent_color link_child_color 2
							4: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane dash_step dash_length link_parent_color link_child_color 3
						)
					)
					else	
					(
						case link_display_mode of
						(
							1: b_plane = drawDIagonalConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane 1 1 link_parent_color link_child_color
							2: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane 1 1 link_parent_color link_child_color 1
							3: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane 1 1 link_parent_color link_child_color 2
							4: b_plane = drawConnection (Point2 from_middle_x from_middle_y) (Point2 to_middle_x to_middle_y) b_plane 1 1 link_parent_color link_child_color 3
						)	
					)	
				)		
			)		
		)
		)
		for n = 1 to TreeMatoGraph_Nodes.count do
		(
			if n != except then 
				b_plane = drawNode n b_plane full_draw
		)
	b_plane 
	)	
	
	fn NodeHitTest pos =
	(
		hit_node = 0
		for n = 1 to TreeMatoGraph_Nodes.count do
			if pos.x >= TreeMatoGraph_Nodes[n].XPos+origin_offset.x and pos.y >= TreeMatoGraph_Nodes[n].YPos+origin_offset.y and pos.x <= TreeMatoGraph_Nodes[n].XPos+origin_offset.x+TreeMatoGraph_Nodes[n].Width and pos.y <= TreeMatoGraph_Nodes[n].YPos+ TreeMatoGraph_Nodes[n].Height+origin_offset.y then
			(
				hit_node = n
				exit
			)
	hit_node		
	)
	
	fn findClosestParent index=
	(
		parent_index = index
		go_on = true
		while go_on do
		(
			parent_index = collect_nodes[parent_index][3]
			try(check_class = superclassof collect_nodes[parent_index][1].object )catch(check_class = 0)
			if check_class == 0 then try(check_class = superclassof collect_nodes[parent_index][1])catch(check_class = 0)
			if check_class == Material or check_class == TextureMap then go_on = false
		)
	parent_index	
	)
	
	fn collecLightsState = 
	(
		lights_state = #()
		lights_array = for l in Lights where classof l != TargetObject collect l
		for l in lights_array do 
		(
			append lights_state l.on
			l.on = false
		)
	)
	
	fn restoreLightsState = 
	(
		for l = 1 to lights_array.count do 
		(
			lights_array[l].on = lights_state[l]
		)
	)

	fn storeAppData node_record =
	(
		app_data_start = 20020515
		setAppData node_record.MaxNode (app_data_start+1) (node_record.Xpos as string)
		setAppData node_record.MaxNode (app_data_start+2) (node_record.YPos as string)
	)

	fn restoreAppData node_record =
	(
		app_data_start = 20020515
		New_Pos = Point2 0.0 0.0
		try
		(
		new_x = Execute( getAppData node_record.MaxNode (app_data_start+1))
		new_y = Execute( getAppData node_record.MaxNode (app_data_start+2))
		if new_x != undefined then New_Pos.x = new_x
		if new_y != undefined then New_Pos.y = new_y
		)
		catch()
		New_pos
	)
	
	fn storeTreePosition =
	(
		for i in TreeMatoGraph_Nodes do
		(
			storeAppData i
		)
	)
	
	fn restoreTreePosition =
	(
		for i in TreeMatoGraph_Nodes do
		(
			read_record = restoreAppData i
			if read_record != [0.0,0.0] then 
			(
				i.XPos = read_record.x
				i.YPos = read_record.y
			)
		)
	)
	
	fn SMIV switch =
	(
		for i in TreeMatoGraph_Nodes do
		(
			if superclassof i.MaxNode == Material then showTextureMap i.MaxNode switch
		)
	)
	
	
	fn CreateNewRoot mat =
	(
		TreeMatoGraph_Nodes = #()
		TreeMatoGraph_Links = #()	
		try
		(
			TreeMatoGraph_Undo_Buffer = copy TreeMatoGraph_Currently_Edited_Tree 
			TreeMatoGraph_Undo_Buffer.name = TreeMatoGraph_Currently_Edited_Tree.name
		)catch()	
		TreeMatoGraph_Currently_Edited_Tree = mat
		map_sample = TreeMatoGraph_rollout.TMG_renderMaterial mat (thumb_width*1.5) (thumb_height*1.5)
		try(mat_color = mat.diffuse)catch(mat_color = color 210 210 210)
			
		new_node = tree_node()
		new_node.Name = mat.name
		new_node.XPos = 300-thumb_width
		new_node.YPos = 2
		new_node.Width = thumb_width*1.5
		new_node.Height = thumb_height*1.5
		new_node.ShadedColor = mat_color
		new_node.Thumbnail = map_sample
		new_node.MaxNode = mat
		new_node.Row = 1 
		new_node.Column = 1
		new_node.Selected = False
		append TreeMatoGraph_Nodes new_node			

		
		copy graph_bmp back_plane
		TreeMatoGraph_rollout.buildTree TreeMatoGraph_Currently_Edited_Tree 
		TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
		TreeMatoGraph_rollout.displayNodeNames()	
		TreeMatoGraph_rollout.UiLog ("Created New "+ (classof mat as string) +" Material ["+mat.name+"]") 2
	)
	
	fn collectMapSlots map =
	(
		temp_map = noise()
		props = getPropNames map
		map_slots = #()		
		if superclassof map == TextureMap then
		(
			case (classof map) of
			(
				Bricks: ( 
					append map_slots #("Bricks - Cannot Set", ".map", "Map")
					append map_slots #("Mortar - Cannot Set", ".map", "Map")
				)
				
				CompositeTexturemap: ( 
					for i = 1 to map.mapList.count do 
					(
						append map_slots #( ("Map "+i as string), ".mapList[" + i as string +"]", "Map" ) 
					)
				)	
				
				default: (
					for p in props do
					(
						try
						(
						if superclassof (getProperty map p) == TextureMap then
						(
							append map_slots #(p, "Property", "Map" )
						)	
						else
						(
							if p != #coords and p != #output then
							(
								success = false
								try
								(
									setProperty map p temp_map
									success = true
									setProperty map p undefined
								)
								catch()
								if success then append map_slots #(p, "Property", "Map")
							)
						)
						)catch()
					)
					)--end default
				)--end case	
		)
		
		if superclassof map == Material then
		(
			case (classof map) of
			(
			
			Advanced_Lighting_Override:	(
										map_slots = #(
										#(#baseMaterial, "Property", "Material"))
								)
			Blend:	(
										map_slots = #(
										#(#Map1, "Property", "Material"),
										#(#Map2, "Property", "Material"),
										#(#Mask, "Property", "Map")
										)
								)	
			compositematerial:	(
										map_slots = #(
										#(#baseMaterial, "Property", "Material"))
										for i = 2 to map.materialList.count do 
										(
											append map_slots #( ("Mat "+(i-1) as string), ".materialList[" + i as string +"]", "Material" ) 
										)										
								)	
			DoubleSided:		(
										map_slots = #(
										#(#Material1, "Property", "Material"),
										#(#Material2, "Property", "Material")
										)
								)		
			Lightscape_Mtl:		(
										map_slots = #(
										#(#baseMaterial, "Property", "Material")
										)
								)	
			MatteShadow:		(
										map_slots = #(
										#(#map, "Property", "Map")
										)
								)	
/*			Morphermaterial:		(
										map_slots = #(
										#(#base, "Property", "Material"))
										for i = 0 to 99 do 
										(
											append map_slots #( ("Mat_"+i as string), ".mat_" + i as string , "Material" ) 
										)	
								)				
*/
											
			Multimaterial:		(
										map_slots = #()
										for i = 1 to map.materialList.count do 
										(
											append map_slots #( ("Mat "+i as string), ".materialList[" + i as string +"]", "Material" ) 
										)										
								)				
			InkNPaint: 			(
									map_slots = #(
										#(#mat1, "Property", "Material"),
										#(#bump_map, "Property", "Map"),
										#(#dsp_map, "Property", "Map"),
										#(#paint_map, "Property", "Map"),
										#(#spec_map, "Property", "Map"),
										#(#shade_color_map, "Property", "Map"),
										#(#width_map, "Property", "Map"),
										#(#out_ink_map, "Property", "Map"),
										#(#smgroup_edge_map, "Property", "Map"),
										#(#matid_map, "Property", "Map"),
										#(#edge_ang_ink_map, "Property", "Map"),
										#(#self_overlap_ink_map, "Property", "Map"),
										#(#self_underlap_ink_map, "Property", "Map")
									)
								)
			Shell_Material:			(
									map_slots = #(
										#(#originalMaterial, "Property", "Material"),
										#(#bakedMaterial, "Property", "Material")
										)
								)			
			Shellac:			(
									map_slots = #(
										#(#shellacMtl1, "Property", "Material"),
										#(#shellacMtl2, "Property", "Material")
										)
								)
			TopBottom:			(
									map_slots = #(
										#(#topMaterial, "Property", "Material"),
										#(#bottomMaterial, "Property", "Material")
										)
								)								
			RaytraceMaterial: 	(
									for p in props do
									(
										if findstring (p as string) "Map" == (p as string).count-2 then append map_slots #(p, "Property", "Map")
									)
								)
			StandardMaterial:	(
									case map.shadertype of
									(
									0:	(
										map_slots = #( 
											#(#AmbientMap, "Property", "Map"),
 											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#DiffuseLevelMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#GlossinessMap,"Property", "Map"), 
											#(#AnisotropyMap, "Property", "Map"), 
											#(#OrientationMap, "Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)
									1:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#GlossinessMap,"Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)
									2:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#GlossinessMap,"Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)										
									3:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#DiffuseLevelMap, "Property", "Map"), 
											#(#DiffuseRoughnessMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#GlossinessMap,"Property", "Map"), 
											#(#AnisotropyMap, "Property", "Map"), 
											#(#OrientationMap, "Property", "Map"), 
											#(#SpecularMap2, "Property", "Map"), 
											#(#SpecularLevelMap2, "Property", "Map"), 
											#(#GlossinessMap2, "Property", "Map"), 
											#(#AnisotropyMap2, "Property", "Map"), 
											#(#OrientationMap2, "Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)
									4:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#GlossinessMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#DiffuseLevelMap, "Property", "Map"), 
											#(#DiffuseRoughnessMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)
									5:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#GlossinessMap, "Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)										
									6:	(
										map_slots = #(
											#(#DiffuseMap,"Property", "Map"), 
											#(#GlossinessMap, "Property", "Map"), 
											#(#MetalnessMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)										
									7:	(
										map_slots = #(
											#(#AmbientMap, "Property", "Map"), 
											#(#DiffuseMap, "Property", "Map"), 
											#(#SpecularMap, "Property", "Map"), 
											#(#GlossinessMap, "Property", "Map"), 
											#(#SpecularLevelMap, "Property", "Map"), 
											#(#SelfIllumMap, "Property", "Map"), 
											#(#OpacityMap, "Property", "Map"), 
											#(#FilterMap, "Property", "Map"), 
											#(#DiffuseLevel, "Property", "Map"), 
											#(#TranslucentColor, "Property", "Map"), 
											#(#BumpMap, "Property", "Map"), 
											#(#ReflectionMap, "Property", "Map"), 
											#(#RefractionMap, "Property", "Map"), 
											#(#DisplacementMap, "Property", "Map") )
										)										
									)					
								)--end standard
			)--end case
		)--end if
	map_slots 
	)	
	
	fn TMG_renderMaterial mat size_x size_y=
	(
		disablesceneredraw()
		with undo off
		(
		UiLog ("Rendering Sample of " + mat.name) 6
		collecLightsState()
		om1 = omniLight pos:[99900.0,99900.0,100300.0] multiplier:1.5
		om2 = omniLight pos:[100100.0,100000.0,99900.0] multiplier:2.0
		
		s = sphere pos:[100000.0,100000.0,100000.0] mapcoords:true 
		s.material = mat
--		addmodifier s (materialModifier materialId:1)		
		t_o = targetObject pos:[100000.0,100000.0,100000.0]
		c = TargetCamera target:t_o
		c.pos = [100000.0,100000.0-68,100020.0] 
		map_sample = render camera:c outputsize:[size_x,size_y] vfb:false
		delete c
		delete s
		delete om1
		delete om2
		restoreLightsState()
		)
		enablesceneredraw()
	map_sample 
	)
	

	fn findNextColInRow row col =
	(
		for c = col to 50 do
		(
			if display_map_array[(row-1)*50+c] == false then 
			(
				return c
			)
		)
	50	
	)	
	
	--Experimental WIP Arrange code.	
	fn autoArrangeLoose =
	(
		rows_array = #{}
		for r = 1 to 10 do
		(
			rows_array[r] = False
			for i = 1 to TreeMatoGraph_Nodes.count do
			(
				if TreeMatoGraph_Nodes[i].Row == r then
				(
					rows_array[r] = True
				)
			)	
		)
		if rows_array.count > 0 then
			row_delta = graph_height / ((rows_array as array).count)
		if row_delta < thumb_vspacing then column_delta = thumb_vspacing
		h_offset = thumb_hspacing
		for r in rows_array do
		(
			current_row = #()
			for i = 1 to TreeMatoGraph_Nodes.count do
			(
				if TreeMatoGraph_Nodes[i].Row == r then
				(
					append current_row r
				)
			)	
			if current_row.count > 0 then
			(
				column_delta = graph_width / (current_row.count+1)
				if column_delta < thumb_hspacing then column_delta = thumb_hspacing
				for i = 1 to current_row.count do
				(
					TreeMatoGraph_Nodes[current_row[i]].XPos = i * column_delta + h_offset
					TreeMatoGraph_Nodes[current_row[i]].YPos = (r-1) * row_delta 
					
				)	
			)	
		h_offset += thumb_hspacing	
		)--end r loop
		UiLog "Ready." 2
	)	
	
	fn buildTree mat =
	(
		copy graph_bmp back_plane
		TreeMatoGraph_Currently_Edited_Tree = mat
		UiLog "Building Material Tree..." 6
		collect_nodes = #(#(mat,mat,0,0,0))
		collect_links = #()
		parent_array = #()
		final_parent_array = #(1)
		cnt = 0
		TreeMatoGraph_Nodes =#()
		TreeMatoGraph_Links =#()
		current_parent = 1
		try(mat_color = mat.diffuse)catch(mat_color = color 210 210 210)
		
		new_node = tree_node()
		new_node.Name = (mat.name+ "[" + (classof mat as string)+"]")
		new_node.XPos = 300-thumb_width
		new_node.YPos = 2
		new_node.Width = thumb_width*1.5
		new_node.Height = thumb_height*1.5
		new_node.ShadedColor = mat_color
		
		isMissingMaps = checkForMissingMaps ()
		if not isMissingMaps then
		(
			map_sample = TMG_renderMaterial mat (thumb_width*1.5) (thumb_height*1.5)
		)
		else
		(
			map_sample = bitmap (thumb_width*1.5) (thumb_height*1.5)
		)
		
		new_node.Thumbnail = map_sample
		new_node.MaxNode = mat
		new_node.Row = 1 
		new_node.Column = 1
		new_node.Selected = False
		append TreeMatoGraph_Nodes new_node
		
		current_depth = 1
		while cnt < collect_nodes.count do
		(
			cnt +=1
			current_cnt = cnt
			try(check_class = superclassof collect_nodes[cnt][1].object)catch(check_class = undefined)
			if check_class == Materials or check_class == TextureMap then current_parent = collect_nodes[cnt][1]
			for m = 1 to collect_nodes[cnt][1].numsubs do
			(
				
				append collect_nodes #(collect_nodes[cnt][1][m], current_parent, cnt, 0,0)
				collect_nodes[cnt][4]=current_cnt
				collect_nodes[cnt][5]=current_cnt+m
			)--end m loop
		)--end while
		
		already_used_array = #()
		already_used_index = #()
		for i = 2 to collect_nodes.count do
		(	
			try
			(
			if superclassof collect_nodes[i][1].object == TextureMap then
			(
				final_parent_index = findClosestParent i
				parent_index = findItem final_parent_array final_parent_index 
				check_instances = findItem already_used_array collect_nodes[i][1].object
				if check_instances == 0 then
				(
					map_name = collect_nodes[i][1].object.name+ "[" + (classof collect_nodes[i][1].object as string)+"]"
					UiLog ("Rendering Sample of " + collect_nodes[i][1].object.name) 6
					map_sample = renderMap collect_nodes[i][1].object scale:100.0 size:[thumb_width,thumb_height] filter:true
					
					new_node = tree_node()
					new_node.Name = map_name
					new_node.XPos = TreeMatoGraph_Nodes.count*thumb_width
					new_node.YPos = thumb_height+parent_index*(thumb_height+thumb_vspacing)
					new_node.Width = thumb_width
					new_node.Height = thumb_height
					new_node.ShadedColor = (color 200 255 200)
					new_node.Thumbnail = map_sample
					new_node.MaxNode = collect_nodes[i][1].object
					new_node.Row = 1 
					new_node.Column = 1
					new_node.Selected = False
					append TreeMatoGraph_Nodes new_node
					
					append TreeMatoGraph_Links #(parent_index,TreeMatoGraph_Nodes.count)
					append final_parent_array i
					append already_used_array collect_nodes[i][1].object
					append already_used_index TreeMatoGraph_Nodes.count
				)
				else	
				(
					append TreeMatoGraph_Links #(parent_index,already_used_index[check_instances])
				)
			)
			)catch()
			try
			(
			if superclassof collect_nodes[i][1]== Material then
			(
				final_parent_index = findClosestParent i
				parent_index = findItem final_parent_array final_parent_index 
				check_instances = findItem already_used_array collect_nodes[i][1]
				if check_instances == 0 then
				(
					map_name = collect_nodes[i][1].name+ "[" + (classof collect_nodes[i][1] as string)+"]"
					UiLog ("Rendering Sample of " + collect_nodes[i][1].name) 6
					try(mat_color = collect_nodes[i][1].diffuse)catch(mat_color = color 210 210 210)
					
					

					new_node = tree_node()
					new_node.Name = map_name
					new_node.XPos = TreeMatoGraph_Nodes.count*thumb_width
					new_node.YPos = thumb_height+parent_index*(thumb_height+thumb_vspacing)
					new_node.Width = thumb_width
					new_node.Height = thumb_height
					new_node.ShadedColor = (color 200 255 200)
					isMissingMaps = checkForMissingMaps ()
					if not isMissingMaps then
					(
						map_sample = TMG_renderMaterial collect_nodes[i][1] (thumb_width) (thumb_height)
					)
					else
					(
						map_sample = bitmap (thumb_width) (thumb_height)
					)
					
					new_node.Thumbnail = map_sample
					new_node.MaxNode = collect_nodes[i][1]
					new_node.Row = 1 
					new_node.Column = 1
					new_node.Selected = False					
					append TreeMatoGraph_Nodes new_node
					
					append TreeMatoGraph_Links #(parent_index,TreeMatoGraph_Nodes.count)
					append final_parent_array i
					append already_used_array collect_nodes[i][1]
					append already_used_index TreeMatoGraph_Nodes.count					
				)
				else	
				(
					append TreeMatoGraph_Links #(parent_index,already_used_index[check_instances])
				)				
			)
			)catch()			
			try
			(
			if superclassof collect_nodes[i][1].object == Material then
			(
				final_parent_index = findClosestParent i
				parent_index = findItem final_parent_array final_parent_index 
				check_instances = findItem already_used_array collect_nodes[i][1].object
				if check_instances == 0 then
				(
					map_name = collect_nodes[i][1].object.name + "[" + (classof collect_nodes[i][1].object as string)+"]"
					UiLog ("Rendering Sample of " + collect_nodes[i][1].object.name) 6
					map_sample = TMG_renderMaterial collect_nodes[i][1].object  (thumb_width) (thumb_height)
					try(mat_color =collect_nodes[i][1].object.diffuse)catch(mat_color = color 210 210 210)
					
					new_node = tree_node()
					new_node.Name = map_name
					new_node.XPos = TreeMatoGraph_Nodes.count*thumb_width
					new_node.YPos = thumb_height+parent_index*(thumb_height+thumb_vspacing)
					new_node.Width = thumb_width
					new_node.Height = thumb_height
					new_node.ShadedColor = (color 200 255 200)

					isMissingMaps = checkForMissingMaps ()
					if not isMissingMaps then
					(
					map_sample = TMG_renderMaterial collect_nodes[i][1].object  (thumb_width) (thumb_height)
					)
					else
					(
						map_sample = bitmap (thumb_width) (thumb_height)
					)
					
					new_node.Thumbnail = map_sample
					new_node.MaxNode = collect_nodes[i][1].object
					new_node.Row = 1 
					new_node.Column = 1
					new_node.Selected = False					
					append TreeMatoGraph_Nodes new_node
					
					append TreeMatoGraph_Links #(parent_index,TreeMatoGraph_Nodes.count)
					append final_parent_array i
					append already_used_array collect_nodes[i][1].object
					append already_used_index TreeMatoGraph_Nodes.count					
				)
				else	
				(
					append TreeMatoGraph_Links #(parent_index,already_used_index[check_instances])
				)		
			)
			)catch()			
		)--end loop
		
		display_map_array = #()
		for i = 1 to 500 do append display_map_array false
		last_row_used = true
		for i = 1 to TreeMatoGraph_Nodes.count do
		(
			if last_row_used then 
			(
				last_row_used = false 
			)	
			for j = 1 to TreeMatoGraph_Links.count do
			(
				if TreeMatoGraph_Links[j][1] == i then
				(
					parent_row = TreeMatoGraph_Nodes[i].Row+1
					parent_col = TreeMatoGraph_Nodes[i].Column
					current_col = findNextColInRow parent_row parent_col
					TreeMatoGraph_Nodes[TreeMatoGraph_Links[j][2]].Row = parent_row
					TreeMatoGraph_Nodes[TreeMatoGraph_Links[j][2]].Column = current_col
					TreeMatoGraph_Nodes[TreeMatoGraph_Links[j][2]].XPos = thumb_width/2.0 + current_col*(thumb_width+thumb_hspacing) 
					TreeMatoGraph_Nodes[TreeMatoGraph_Links[j][2]].YPos = (parent_row-1)*thumb_height*2.0
					last_row_used = true
					display_map_array[(parent_row-1)*50+current_col] = true
				)
			)
		)--end i
		UiLog "Ready." 2
		--autoArrangeLoose()
	)
	

	fn updateAllThumbnails =
	(
		for i = 1 to TreeMatoGraph_Nodes.count do
		(
			if superclassof TreeMatoGraph_Nodes[i].MaxNode == TextureMap then
			(
				map_sample = renderMap TreeMatoGraph_Nodes[i].MaxNode scale:100.0 size:[TreeMatoGraph_Nodes[i].Width ,TreeMatoGraph_Nodes[i].Height] filter:true
				TreeMatoGraph_Nodes[i].Thumbnail = map_sample 
			)	
			else
			(
				isMissingMaps = checkForMissingMaps()
				if not isMissingMaps then
				(
					map_sample = TMG_renderMaterial TreeMatoGraph_Nodes[i].MaxNode TreeMatoGraph_Nodes[i].Width TreeMatoGraph_Nodes[i].Height
				)
				else
				(
					map_sample = bitmap TreeMatoGraph_Nodes[i].Width TreeMatoGraph_Nodes[i].Height
				)
				TreeMatoGraph_Nodes[i].Thumbnail = map_sample 
			)	
		)	
		graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
		UiLog ("Updated All Thumbnails") 1
		displayNodeNames()
	)
	
	on TreeMatoGraph_rollout mbuttondown pos do
	(
		start_pos = pos
		MButtonPressed = true
		pan_offset = [0.0,0.0]
		current_node = 0
		displayNodeNames()
	)	
	
	on TreeMatoGraph_rollout mbuttonDblClk pos do
	(
		origin_offset = [-40.0,0.0]
		current_node = 0
		copy graph_bmp back_plane
		graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
		displayNodeNames()
	)		
	
	on TreeMatoGraph_rollout mbuttonup pos do
	(
		setSysCur #arrow
		MButtonPressed = false
		start_pos = [0,0]
		pan_offset = [0.0,0.0]
		copy graph_bmp back_plane
		graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
		displayNodeNames()
	)	
	
	fn findParents node =
	(
		dependent_nodes = #()
		for l in TreeMatoGraph_Links do
		(
			if l[2] == node then append dependent_nodes l[1]
		)
	dependent_nodes
	)

	fn findChildren node =
	(
		dependent_nodes = #()
		for l in TreeMatoGraph_Links do
		(
			if l[1] == node then append dependent_nodes l[2]
		)
	dependent_nodes
	)
	
	fn FindMultipleSelection from_pos to_pos = 
	(
	
		selected_count = 0
		if from_pos.x > to_pos.x then
		(
			upper_left_x = to_pos.x
			lower_right_x = from_pos.x
		)	
		else
		(
			upper_left_x = from_pos.x
			lower_right_x = to_pos.x
		)
		if from_pos.y > to_pos.y then
		(
			upper_left_y = to_pos.y
			lower_right_y = from_pos.y
		)	
		else
		(
			upper_left_y = from_pos.y
			lower_right_y = to_pos.y
		)	
		for i = 1 to TreeMatoGraph_Nodes.count do
		(
			obj = TreeMatoGraph_Nodes[i]
			if obj.XPos-obj.Width/2.0+origin_offset.x < lower_right_x and obj.XPos+obj.Width/2.0+origin_offset.x > upper_left_x and obj.YPos-obj.Height/2.0+origin_offset.y < lower_right_y and obj.YPos+obj.Height/2.0+origin_offset.y > upper_left_y then
			(
				TreeMatoGraph_Nodes[i].Selected = True
				selected_count += 1
			)
			else
			(
				TreeMatoGraph_Nodes[i].Selected = False
			)
		)
		selected_count 
	)--end fn
	
	fn UnselectAllNodes =
	(
		for i in TreeMatoGraph_Nodes do i.Selected = False
	)
	
	fn showParentChannels current_node =
	(
			t_map = TreeMatoGraph_Nodes[current_node].MaxNode
			parent_list = findParents current_node
			parent_list = sort parent_list 
			last_parent = 0
			for d in parent_list do
			(
				if last_parent != d then 
				(
					parent_count = 0
					s_map = TreeMatoGraph_Nodes[d].MaxNode
					source_pos_x = TreeMatoGraph_Nodes[d].XPos
					source_pos_y = TreeMatoGraph_Nodes[d].YPos
					source_slots = collectMapSlots s_map
					for i in source_slots do
					(
						if i[2] == "Property" then
						(
							if getProperty s_map i[1] == t_map then 
							(
								type_counter += 1
								parent_count += 1
								Type_Label[type_counter].text = (i[1] as string)
								Type_Label[type_counter].pos.x = source_pos_x + origin_offset.x
								Type_Label[type_counter].pos.y = source_pos_y + parent_count*14 + origin_offset.y + TreeMatoGraph_Nodes[d].Height
							)	
						)	
						else
						(
							try(test_value = execute ("s_map"+i[2]))catch(test_value = undefined)
							if  test_value == t_map then 
							(
								type_counter += 1
								parent_count += 1
								Type_Label[type_counter].text = (i[1] as string)
								Type_Label[type_counter].pos.x = source_pos_x +origin_offset.x
								Type_Label[type_counter].pos.y = source_pos_y + parent_count*14 +origin_offset.y + TreeMatoGraph_Nodes[d].Height
							)--end if	
						)--end else	
					)--end i loop
				)--end if 
				last_parent = d
			)--end d loop
	)--end fn	
	
	fn showChildChannels current_node =
	(
			t_map = TreeMatoGraph_Nodes[current_node].MaxNode
			target_pos_x = TreeMatoGraph_Nodes[current_node].XPos
			target_pos_y = TreeMatoGraph_Nodes[current_node].YPos
			
			children_list = findChildren current_node
			children_list = sort children_list 
			last_child = 0
			for d in children_list do
			(
				if last_child != d then 
				(
					children_count = 0
					s_map = TreeMatoGraph_Nodes[d].MaxNode
					source_pos_x = TreeMatoGraph_Nodes[d].XPos
					source_pos_y = TreeMatoGraph_Nodes[d].YPos
					target_slots = collectMapSlots t_map
					for i in target_slots do
					(
						if i[2] == "Property" then
						(
							if getProperty t_map i[1] == s_map then 
							(
								type_counter += 1
								children_count += 1
								Type_Label[type_counter].text = (i[1] as string)
								Type_Label[type_counter].pos.x = source_pos_x + origin_offset.x
								Type_Label[type_counter].pos.y = source_pos_y + children_count*14 + origin_offset.y + TreeMatoGraph_Nodes[d].Height
							)	
						)	
						else
						(
							try(test_value = execute ("t_map"+i[2]))catch(test_value = undefined)
							if  test_value == s_map then 
							(
								type_counter += 1
								children_count += 1
								Type_Label[type_counter].text = (i[1] as string)
								Type_Label[type_counter].pos.x = source_pos_x +origin_offset.x
								Type_Label[type_counter].pos.y = source_pos_y + children_count*14 +origin_offset.y + TreeMatoGraph_Nodes[d].Height
							)--end if	
						)--end else	
					)--end i loop
				)--end if 
				last_child = d
			)--end d loop
	)--end fn	
	
	
	fn createConnection pos=
	(
		if ConnectMode then
		(
			current_node = NodeHitTest pos
			if current_node > 0 and current_node != connect_source_node then
			(
				target_map = TreeMatoGraph_Nodes[current_node].MaxNode
				source_map = TreeMatoGraph_Nodes[connect_source_node].MaxNode
				cannot_connect = (superclassof source_map == Material and superclassof target_map == TextureMap)
				if cannot_connect then TreeMatoGraph_rollout.UiLog ("Cannot Link Material To Map!") 4
				source_slots = collectMapSlots source_map
				for i in source_slots do
				(
					if i[2] == "Property" then
					(
						if getProperty source_map i[1] == target_map then 
						(
							cannot_connect = true
							TreeMatoGraph_rollout.UiLog ("Cannot Link - Circular Dependency!") 4
						)	
					)	
					else
					(
						try(test_value = execute ("source_map"+i[2]))catch(test_value = undefined)
						if  test_value == target_map then 
						(
							cannot_connect = true
							TreeMatoGraph_rollout.UiLog ("Cannot Link - Circular Dependency!") 4
						)	
					)	
				)
				
				if cannot_connect then 
				(
					destroyDialog TreeMatoGraph_link_map_rollout
				)
				else
				(
					target_slots = collectMapSlots target_map
					if target_slots.count > 0 then
					(
						tmg_pos = getDialogPos TreeMatoGraph_rollout 
						rollout_height = (100+target_slots.count*16)
						if rollout_height > 400 then rollout_height = 400
						createDialog TreeMatoGraph_link_map_rollout 200 rollout_height (TreeMatoGraph_Nodes[current_node].XPos+tmg_pos.x) (TreeMatoGraph_Nodes[current_node].YPos+TreeMatoGraph_Nodes[current_node].Height+20+tmg_pos.y  )  --modal:true
						TreeMatoGraph_link_map_rollout.label_source.text = (source_map.name + "[" + (classof source_map) as string +"]" )
						TreeMatoGraph_link_map_rollout.label_target.text = (target_map.name + "[" + (classof target_map) as string +"]" )
						local names_array = #()
						for i in target_slots do
						(
							if i[2] == "Property" then 
							(
								current_value = getProperty target_map i[1]
							)	
							else
							(
								try(current_value = execute("target_map"+ i[2]))catch(current_value=undefined)
							)	
							if current_value == undefined then
								current_value_string = ""
							else
								try(current_value_string = " - "+current_value.name)catch(current_value_string = "")
							append names_array ((i[1] as string) + current_value_string)
						)	
						
						TreeMatoGraph_link_map_rollout.map_list.items = names_array
					)
				)
			)
			ConnectMode = false
		)	
	)--end connect function
	
	fn destroyConnection pos=
	(
		if DisconnectMode then
		(
			store_old_positions()
			current_node = NodeHitTest pos
			if current_node > 0 and current_node != connect_source_node then
			(
				target_map = TreeMatoGraph_Nodes[current_node].MaxNode
				source_map = TreeMatoGraph_Nodes[connect_source_node].MaxNode
				target_slots = collectMapSlots target_map
				linked_slots = #()
				for t = 1 to target_slots.count do
				(
					if target_slots[t][2] == "Property" then
					(
						if getProperty target_map target_slots[t][1] == source_map then 
						(
							append linked_slots target_slots[t]
						)
					)
					else
					(	
						try(test_value = execute ("target_map"+target_slots[t][2]))catch(test_value = undefined)
						if test_value == source_map then append linked_slots target_slots[t]
					)
				)
				if linked_slots.count > 0 then 
				(
					tmg_pos = getDialogPos TreeMatoGraph_rollout 
					local names_array = #()
					for i in linked_slots do
					(
						--if i[2] == "Property" then 
						append names_array (i[1] as string)
					)	
					rollout_height = (100+names_array.count*16)
					if rollout_height > 400 then rollout_height = 400
					createDialog TreeMatoGraph_unlink_map_rollout 200 rollout_height (TreeMatoGraph_Nodes[current_node].XPos+tmg_pos.x) (TreeMatoGraph_Nodes[current_node].YPos+TreeMatoGraph_Nodes[current_node].Height+20+tmg_pos.y  ) --modal:true
					TreeMatoGraph_unlink_map_rollout.label_source.text = (source_map.name + "[" + (classof source_map) as string +"]" )
					TreeMatoGraph_unlink_map_rollout.label_target.text = (target_map.name + "[" + (classof target_map) as string +"]" )
					
					TreeMatoGraph_unlink_map_rollout.map_list.items = names_array
				)
				else
				(
					UiLog "Nothing to Unlink." 4
				)	
			)
			DisconnectMode = false
		)	
	)--end connect function
	
	on rescan_tree pressed do
	(
		if TreeMatoGraph_Currently_Edited_Tree != undefined then 
		(
			store_old_positions()
			buildTree TreeMatoGraph_Currently_Edited_Tree 
			restoreTreePosition()	
			restore_old_positions()		
			graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1
			UiLog "Reloaded Current Material Tree." 2
			
			displayNodeNames()
		)	
	)
	
	on TreeMatoGraph_rollout lbuttonup pos do
	(
		if ConnectMode then createConnection pos
		if DisconnectMode then destroyConnection pos
		tooltip_label.pos = [1000,1000]
		setSysCur #arrow
		LButtonPressed =false
		start_pos = [0,0]
		if multiple_selection_mode then
		(
			multiple_selection_end = pos
			multiple_selection_mode = false
			multiple_selection_count = FindMultipleSelection multiple_selection_start multiple_selection_end
			UiLog ("Selected "+multiple_selection_count as string +" Node(s).") 2
		)	
		copy graph_bmp back_plane
		graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
		displayNodeNames()
		move_mode.checked = true
		link_mode.checked = false
		unlink_mode.checked = false		
	)	
	
	on TreeMatoGraph_rollout rbuttonDblClk pos do
	(

		current_node = NodeHitTest pos
		if current_node == 0 then
		(
			move_mode.checked = true
			link_mode.checked = false
			del_mode.checked = false
			unlink_mode.checked = false
		)
	)
	
	on TreeMatoGraph_rollout rbuttonup pos do
	(
			if keyboard.controlPressed then
			(	
				DisconnectMode = true
				destroyConnection pos
			)	
			else	
			(
				createConnection pos
			)	
			tooltip_label.pos = [1000,1000]
			setSysCur #arrow
			RButtonPressed =false
			start_pos = [0,0]
			copy graph_bmp back_plane
			graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
			displayNodeNames()
			move_mode.checked = true
			link_mode.checked = false
			unlink_mode.checked = false			
	)	
	
	
	on TreeMatoGraph_rollout lbuttondown pos do
	(
		old_pos = pos
		current_node = NodeHitTest pos

		if current_node > 0 and LButtonPressed == false and link_mode.checked  then
		(
			LButtonPressed = true
			ConnectMode = true
			connect_source_node = current_node
			copy graph_bmp background_plane 
			background_plane = drawAllNodes background_plane 0 1 1 1
			graph_bitmap.bitmap = background_plane 
		)
		
		if current_node > 0 and LButtonPressed == false and move_mode.checked  then
		(
			LButtonPressed = true
			start_move_pos = pos
--			clip_offset = pos-(Point2 TreeMatoGraph_Nodes[current_node].XPos TreeMatoGraph_Nodes[current_node].YPos  )
			copy graph_bmp background_plane 
			background_plane = drawAllNodes background_plane current_node 1000 1000 1
			copy background_plane back_plane
			back_plane = drawMoveNodeLines back_plane current_node 10 5	link_parent_color link_child_color
			graph_bitmap.bitmap = drawNode current_node back_plane 1
		)
		
		if current_node > 0 and LButtonPressed == false and unlink_mode.checked  then
		(
			LButtonPressed = true
			DisconnectMode = true
			connect_source_node = current_node
			copy graph_bmp background_plane 
			background_plane = drawAllNodes background_plane 0 1 1 1
			graph_bitmap.bitmap = background_plane 
		)

		if current_node > 0 then 
		(
			node_to_rename = TreeMatoGraph_Nodes[current_node].MaxNode
			map_edit_name.text = node_to_rename.name
		)	
		
		if current_node == 0 then 
		(
			multiple_selection_start = pos
			multiple_selection_mode = true
			UnselectAllNodes()
			copy graph_bmp background_plane 
			background_plane = drawAllNodes background_plane 0 1 1 1
			graph_bitmap.bitmap = background_plane 
		)	
	)	
	
	on map_edit_name changed txt do
	(
		if node_to_rename != undefined then node_to_rename.name= txt
		displayNodeNames()
	)


	
	on TreeMatoGraph_rollout rbuttondown pos do
	(
		old_pos = pos
		current_node = NodeHitTest pos

		if current_node > 0 and RButtonPressed == false then
		(
			RButtonPressed = true
			ConnectMode = true
			connect_source_node = current_node
			copy graph_bmp background_plane 
			background_plane = drawAllNodes background_plane 0 1 1 1
			graph_bitmap.bitmap = background_plane 
			if keyboard.controlPressed then
			(
				move_mode.checked = false
				link_mode.checked = false
				unlink_mode.checked = true			
			)
			else
			(
				move_mode.checked = false
				link_mode.checked = true
				unlink_mode.checked = false			
			)	
		)
	)	
	
	fn drawLinkingLines pos =
	(
		copy background_plane back_plane
		back_plane = drawConnectNodeLine back_plane connect_source_node (pos-origin_offset) 4 2 link_link_color [0,0,0]
		graph_bitmap.bitmap = back_plane 
		displayNodeNames()	
	)
	
	fn drawUnlinkLines pos = 
	(
		copy background_plane back_plane
		back_plane = drawConnectNodeLine back_plane connect_source_node (pos-origin_offset) 4 2 link_unlink_color [0,0,0]
		graph_bitmap.bitmap = back_plane 
		displayNodeNames()		
	)
	
	on TreeMatoGraph_rollout mousemove pos do
	(
		st = timestamp()
		
		if (LButtonPressed and link_mode.checked) then drawLinkingLines pos
		if (RbuttonPressed and not keyboard.controlPressed) then 
		(
			move_mode.checked = false
			link_mode.checked = true
			unlink_mode.checked = false
			drawLinkingLines pos
		)	
		
		if LButtonPressed and unlink_mode.checked then 
		(
			drawUnlinkLines pos
		)
		else
		(
			if RButtonPressed and keyboard.controlPressed then 
			(
				move_mode.checked = false
				link_mode.checked = false
				unlink_mode.checked = true
				drawUnlinkLines pos			
			)		
		)	
		
		if multiple_selection_mode then
		(
			multiple_selection_end = pos
			copy background_plane back_plane
			drawSelectionWindow back_plane multiple_selection_start multiple_selection_end
			graph_bitmap.bitmap = back_plane 
			displayNodeNames()				
		)

		
		if LButtonPressed and move_mode.checked  then
		(
			setSysCur #move
			if new_pos != old_pos then
			(
				if multiple_selection_count == 0 then
				(
					TreeMatoGraph_Nodes[current_node].XPos += (pos-start_move_pos ).x
					TreeMatoGraph_Nodes[current_node].YPos += (pos-start_move_pos ).y
					copy background_plane back_plane
					back_plane = drawMoveNodeLines back_plane current_node 10 5	link_parent_color link_child_color
					back_plane = drawNode current_node back_plane 1
					graph_bitmap.bitmap = drawBorder current_node link_parent_color back_plane 
				)
				else
				(
					for i in TreeMatoGraph_Nodes do
					(
						if i.Selected then
						(
							i.XPos += (pos-start_move_pos).x
							i.YPos += (pos-start_move_pos ).y
						)	
					)
					copy graph_bmp back_plane
					graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1
				)	
				storeTreePosition()
				start_move_pos = pos
			)	
		)
		else
		(
			if MButtonPressed then
			(
				pan_offset = pos - start_pos
				setSysCur #move
				if length pan_offset != 0 then
				(
					origin_offset += pan_offset
					start_pos = pos
					copy graph_bmp back_plane
					graph_bitmap.bitmap = drawAllNodes back_plane 0 10 5 0
				)	
			)			
			else
			(
				If not RbuttonPressed and not LbuttonPressed then
				(
					type_counter = 0
					current_node = NodeHitTest pos
					if current_node > 0  then
					(
						if current_node != current_highlight then
						(
							current_highlight = current_node
							copy graph_bitmap.bitmap back_plane
							back_plane = drawMoveNodeLines back_plane current_node 10 5	link_parent_color link_child_color
							back_plane = drawNode current_node back_plane 1
							graph_bitmap.bitmap = drawBorder current_node link_parent_color back_plane 
							displayNodeNames()	
							if show_parent_channel.checked then	showParentChannels current_node
							if show_child_channel.checked then showChildChannels current_node
						)	
					)	
					else
					(
						if current_highlight != 0 then
						(
							current_highlight = 0 
							copy graph_bmp background_plane
							graph_bitmap.bitmap = drawAllNodes background_plane 0 1 1 1
							clearNodeTypes()
							displayNodeNames()	
						)
					)	
				)
			)	
		)	
			
		et = timestamp()
		pushprompt ("TMG MOVE in "+ ((et-st) as string) + " ms.")
	)--end on mousemove
	
	on show_node_names changed state do displayNodeNames()
	on show_node_class changed state do displayNodeNames()
	
	
	
	fn getNewMapNode pos =
	(
--		new_map = materialBrowseDlg #maps
		new_map = copy (materialBrowseDlg #maps)
		if new_map != undefined then
		(
			map_sample = renderMap new_map scale:100.0 size:[thumb_width,thumb_height] filter:true
			new_node = tree_node()
			new_node.Name = new_map.name
			new_node.XPos = pos.x-thumb_width/2
			new_node.YPos = pos.y-thumb_height/2
			new_node.Width = thumb_width
			new_node.Height = thumb_height
			new_node.ShadedColor = (color 200 255 200)
			new_node.Thumbnail = map_sample
			new_node.MaxNode = new_map
			new_node.Row = 0 
			new_node.Column = 0
			new_node.Selected = False
			append TreeMatoGraph_Nodes new_node		
			UiLog ("Created New Map : "+new_map.name+" ["+ classof new_map as string +"]" ) 2
		)
		else
		(
			UiLog "New Map Node Creation Canceled." 4
		)
	)
	
	fn getNewMatNode pos =
	(
		new_mat = copy (materialBrowseDlg #mats)
		if new_mat != undefined then
		(
			map_sample = TMG_renderMaterial new_mat thumb_width thumb_height
			new_node = tree_node()
			new_node.Name = new_mat.name
			new_node.XPos = pos.x-thumb_width/2
			new_node.YPos = pos.y-thumb_height/2
			new_node.Width = thumb_width
			new_node.Height = thumb_height
			try(new_node.ShadedColor = new_mat.diffuse)catch(new_node.ShadedColor = (color 200 255 200))
			new_node.Thumbnail = map_sample
			new_node.MaxNode = new_mat
			new_node.Row = 0 
			new_node.Column = 0
			new_node.Selected = False			
			append TreeMatoGraph_Nodes new_node		
			UiLog ("Created New Mat : "+new_mat.name+" ["+ classof new_mat as string +"]" ) 2
		)
		else
		(
			UiLog "New Mat Node Creation Canceled." 4
		)
	)	
	
	on TreeMatoGraph_rollout lbuttonDblClk pos do
	(
		local hit_node = NodeHitTest pos
		if hit_node > 0 then 
		(
			max mtledit
			setMeditMaterial 1 TreeMatoGraph_Nodes[hit_node].MaxNode
			medit.setActiveMtlSlot 1 true
			if superclassof TreeMatoGraph_Nodes[hit_node].MaxNode == TextureMap then
			(
				map_sample = renderMap TreeMatoGraph_Nodes[hit_node].MaxNode scale:100.0 size:[TreeMatoGraph_Nodes[hit_node].Width,TreeMatoGraph_Nodes[hit_node].Height] filter:true
				TreeMatoGraph_Nodes[hit_node].Thumbnail = map_sample 
				graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
				UiLog ("Updated Thumbnail of [" + TreeMatoGraph_Nodes[hit_node].MaxNode.name+"]" ) 2
				displayNodeNames()
			)	
			else
			(
				map_sample = TMG_renderMaterial TreeMatoGraph_Nodes[hit_node].MaxNode TreeMatoGraph_Nodes[hit_node].Width TreeMatoGraph_Nodes[hit_node].Height 
				TreeMatoGraph_Nodes[hit_node].Thumbnail = map_sample 
				graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
				UiLog ("Updated Thumbnail of [" + TreeMatoGraph_Nodes[hit_node].MaxNode.name+"]" ) 2
				displayNodeNames()
			)
		)	
		else
		(
			if TreeMatoGraph_Nodes.count == 0 then
			(
				mat = copy (materialBrowseDlg #mats)
				if mat != undefined then
					CreateNewRoot mat 			
				else
					UiLog "Material Tree Root Creation Failed!" 4		
			)
			else
			(
				getNewMapNode (pos-origin_offset)
				storeTreePosition()
				graph_bitmap.bitmap = drawAllNodes back_plane 0 1 1 1	
				displayNodeNames()
			)	
		)
	)	
)

try(destroyDialog TreeMatoGraph_rollout)catch()
gc light:true

createDialog TreeMatoGraph_rollout graph_width (graph_height+42) menu:TMG_RCMenu  style:#(#style_titlebar,#style_sysmenu)  --#style_resizing
registerViewWindow TreeMatoGraph_rollout
TreeMatoGraph_rollout.DefineThumbLabel()
TreeMatoGraph_rollout.DefineTypeLabel()
TreeMatoGraph_rollout.defineColArray()
restoreIniFile()
if TreeMatoGraph_Nodes == undefined then TreeMatoGraph_Nodes = #()
if TreeMatoGraph_Links == undefined then TreeMatoGraph_Links = #()
TreeMatoGraph_rollout.UiLog ("Ready.") 2

if selection.count == 1 then 
(
	if $.material != undefined then
	(
		TreeMatoGraph_rollout.buildTree $.material
		TreeMatoGraph_rollout.restoreTreePosition()
	)	
)	
TreeMatoGraph_rollout.graph_bitmap.bitmap = TreeMatoGraph_rollout.drawAllNodes back_plane 0 1 1 1
TreeMatoGraph_rollout.displayNodeNames()

)--end script